@@ -24,6 +24,7 @@
 */
 
 #include <assert.h>
+#include <time.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <CL/opencl.h>
@@ -51,13 +52,20 @@ main (int argc, char **argv)
   int i, err, spir, spirv, poclbin;
 
   cl_context context = NULL;
-  cl_device_id device = NULL;
   cl_platform_id platform = NULL;
-  cl_command_queue queue = NULL;
   cl_program program = NULL;
+  cl_uint num_devices = 0;
+  cl_device_id *devices = NULL;
+  cl_command_queue *queues = NULL;
+
+  srand ((unsigned int)time (NULL));
+
+  err = poclu_get_multiple_devices (&platform, &context, &num_devices,
+                                    &devices, &queues);
 
-  err = poclu_get_any_device2 (&context, &device, &queue, &platform);
   CHECK_OPENCL_ERROR_IN ("clCreateContext");
+  cl_device_id first_dev = devices[0];
+  cl_command_queue first_queue = queues[0];
 
   spir = (argc > 1 && argv[1][0] == 's');
   spirv = (argc > 1 && argv[1][0] == 'v');
@@ -65,8 +73,10 @@ main (int argc, char **argv)
   const char *explicit_binary_path = (poclbin && (argc > 2)) ? argv[2] : NULL;
 
   const char *basename = "example1";
-  err = poclu_load_program (context, device, basename, spir, spirv, poclbin,
-                            explicit_binary_path, NULL, &program);
+  err = poclu_load_program_multidev (context, devices, num_devices, basename,
+                                     spir, spirv, poclbin,
+                                     explicit_binary_path, NULL, &program);
+
   if (err != CL_SUCCESS)
     goto FINISH;
 
@@ -89,8 +99,8 @@ main (int argc, char **argv)
 
   err = 0;
 
-  if (exec_dot_product_kernel (context, device, queue, program, N, srcA, srcB,
-                               dst))
+  if (exec_dot_product_kernel (context, first_dev, first_queue, program, N,
+                               srcA, srcB, dst))
     {
       printf ("Error running the tests\n");
       err = 1;
@@ -118,9 +128,12 @@ main (int argc, char **argv)
 
 FINISH:
   CHECK_CL_ERROR (clReleaseProgram (program));
-  CHECK_CL_ERROR (clReleaseCommandQueue (queue));
-  CHECK_CL_ERROR (clUnloadPlatformCompiler (platform));
+  for (i = 0; i < num_devices; ++i)
+    {
+      CHECK_CL_ERROR (clReleaseCommandQueue (queues[i]));
+    }
   CHECK_CL_ERROR (clReleaseContext (context));
+  CHECK_CL_ERROR (clUnloadPlatformCompiler (platform));
   free (srcA);
   free (srcB);
   free (dst);
