@@ -27,7 +27,7 @@
 #include <CL/opencl.h>
 #include "poclu.h"
 
-#ifdef _MSC_VER
+#ifdef _WIN32
 #  include "vccompat.hpp"
 #endif
 
@@ -39,7 +39,7 @@ int
 main (int argc, char **argv)
 {
   cl_float *input = NULL, *output = NULL;
-  int i, j, err, spir, spirv;
+  int i, j, err, spir, spirv, poclbin;
   cl_mem memobjs[2] = { 0 };
   size_t global_work_size[2] = { 0 };
   size_t local_work_size[2] = { 0 };
@@ -56,9 +56,12 @@ main (int argc, char **argv)
 
   spir = (argc > 1 && argv[1][0] == 's');
   spirv = (argc > 1 && argv[1][0] == 'v');
+  poclbin = (argc > 1 && argv[1][0] == 'b');
+  const char *explicit_binary_path = (poclbin && (argc > 2)) ? argv[2] : NULL;
 
   const char *basename = "example2";
-  err = poclu_load_program (context, device, basename, spir, spirv, &program);
+  err = poclu_load_program (context, device, basename, spir, spirv, poclbin,
+                            explicit_binary_path, NULL, &program);
   if (err != CL_SUCCESS)
     goto ERROR;
 
@@ -66,13 +69,14 @@ main (int argc, char **argv)
   output = (cl_float *) malloc (WIDTH * (HEIGHT + PADDING) * sizeof (cl_float));
 
   srand48(0);
-  for (i = 0; i < HEIGHT; ++i)
+  for (i = 0; i < WIDTH; ++i)
     {
-      for (j = 0; j < WIDTH; ++j)
-      input[i * WIDTH + j] = (cl_float)drand48();
+      for (j = 0; j < HEIGHT; ++j)
+      input[i * HEIGHT + j] = (cl_float)drand48();
+      for (j = 0; j < (HEIGHT + PADDING); ++j)
+      output[i * (HEIGHT + PADDING) + j] = 0.0f;
     }
 
-  queue = clCreateCommandQueue (context, device, 0, NULL);
   CHECK_CL_ERROR2 (err);
 
   memobjs[0] = clCreateBuffer(context,
@@ -134,8 +138,8 @@ ERROR:
   CHECK_CL_ERROR (clReleaseKernel (kernel));
   CHECK_CL_ERROR (clReleaseProgram (program));
   CHECK_CL_ERROR (clReleaseCommandQueue (queue));
-  CHECK_CL_ERROR (clUnloadPlatformCompiler (platform));
   CHECK_CL_ERROR (clReleaseContext (context));
+  CHECK_CL_ERROR (clUnloadPlatformCompiler (platform));
   free (input);
   free (output);
 
