@@ -16,6 +16,8 @@
 #include <CL/opencl.h>
 #include "poclu.h"
 
+#undef SRCDIR
+#include "config.h"
 
 #define GRID_GRANULARITY 2
 
@@ -26,8 +28,17 @@ typedef struct grid_t {
   cl_int ni, nj, nk;            // used size
 } grid_t;
 
+static int initialised = 0;
+static cl_context context;
+static cl_command_queue cmd_queue;
+static cl_program program;
+static cl_kernel kernel;
+
+
+#ifdef _CL_DISABLE_DOUBLE
+#error Scalarwave test requires cl_khr_fp64 support.
+#endif
 
- 
 int 
 exec_scalarwave_kernel(char      const *const program_source, 
                        cl_double       *const phi,
@@ -35,11 +46,6 @@ exec_scalarwave_kernel(char      const *const program_source,
                        cl_double const *const phi_p_p,
                        grid_t    const *const grid)
 { 
-  static int initialised = 0;
-  static cl_context context;
-  static cl_command_queue cmd_queue;
-  static cl_program program;
-  static cl_kernel kernel;
   
   if (!initialised) {
     initialised = 1;
@@ -123,10 +129,6 @@ exec_scalarwave_kernel(char      const *const program_source,
   clReleaseMemObject(mem_phi_p);
   clReleaseMemObject(mem_phi_p_p);
   clReleaseMemObject(mem_grid);
-  /* clReleaseKernel(kernel); */
-  /* clReleaseProgram(program); */
-  /* clReleaseCommandQueue(cmd_queue); */
-  /* clReleaseContext(context); */
  
   return 0;
 }
@@ -145,14 +147,11 @@ static int roundup(int const nx)
   return (nx + GRID_GRANULARITY-1) / GRID_GRANULARITY * GRID_GRANULARITY;
 }
 
-#ifndef SRCDIR
-#  define SRCDIR "."
-#endif
 
 int
 main(void)
 {
-  FILE *const source_file = fopen(SRCDIR "/scalarwave.cl", "r");
+  FILE *const source_file = fopen(SRCDIR "/examples/scalarwave/scalarwave.cl", "r");
   assert(source_file != NULL && "scalarwave.cl not found!");
   
   fseek(source_file, 0, SEEK_END);
@@ -165,8 +164,6 @@ main(void)
   
   fclose(source_file);
 
-
-
   grid_t grid;
   grid.dt = ALPHA/(NX-1);
   grid.dx = grid.dy = grid.dz = 1.0/(NX-1);
@@ -220,7 +217,15 @@ main(void)
     assert(!ierr);
     
   } // for n
-  
+
+  clReleaseKernel (kernel);
+  clReleaseProgram (program);
+  clReleaseCommandQueue (cmd_queue);
+  cl_platform_id pocl;
+  clGetPlatformIDs (1, &pocl, NULL);
+  clReleaseContext (context);
+  clUnloadPlatformCompiler (pocl);
+
   for (int i=0; i<NX; ++i) {
     int const j = i;
     int const k = i;
@@ -231,9 +236,13 @@ main(void)
     
     printf ("phi(%-8g,%-8g,%-8g) = %g\n", x,y,z, phi[ind3d]);
   }
-  
+
   printf ("Done.\n");
 
+  free (phi);
+  free (phi_p);
+  free (phi_p_p);
   free(source);
+
   return 0;
 }
