@@ -40,6 +40,8 @@
 
 #include "config.h"
 
+#include "pocl_export.h"
+
 #include "pocl_context.h"
 
 /* detects restrict, variadic macros etc */
@@ -52,37 +54,62 @@
 #define CL_KHRONOS_VENDOR_ID_POCL 0x10006
 #endif
 
-typedef struct _mem_mapping mem_mapping_t;
 /* represents a single buffer to host memory mapping */
-struct _mem_mapping {
-  void *host_ptr; /* the location of the mapped buffer chunk in the host memory */
-  size_t offset; /* offset to the beginning of the buffer */
+typedef struct mem_mapping
+{
+  void *host_ptr; /* the location of the mapped buffer chunk in the host memory
+                   */
+  size_t offset;  /* offset to the beginning of the buffer */
   size_t size;
-  mem_mapping_t *prev, *next;
-  /* This is required, because two clEnqueueMap() with the same buffer+size+offset,
-     will create two identical mappings in the buffer->mappings LL.
-     Without this flag, both corresponding clEnqUnmap()s will find
-     the same mapping (the first one in mappings LL), which will lead
+  struct mem_mapping *prev, *next;
+
+  /* This is required, because two clEnqueueMap() with the same
+     buffer+size+offset, will create two identical mappings in the
+     buffer->mappings LL. Without this flag, both corresponding clEnqUnmap()s
+     will find the same mapping (the first one in mappings LL), which will lead
      to memory double-free corruption later. */
-  long unmap_requested;
+  int unmap_requested;
+
   cl_map_flags map_flags;
   /* image mapping data */
   size_t origin[3];
   size_t region[3];
   size_t row_pitch;
   size_t slice_pitch;
-};
+} mem_mapping_t;
 
-/* memory identifier: id to point the global memory where memory resides
-                      + pointer to actual data */
-typedef struct _pocl_mem_identifier
+/* memory identifier:  */
+typedef struct pocl_mem_identifier
 {
-  int available; /* ... in this mem objs context */
-  int global_mem_id;
+  /* global-memory-specific pointer
+     to hardware resource that represents memory. This may be anything, but
+     must be non-NULL while the memory is actually allocated, and NULL when
+     it's not */
   void *mem_ptr;
-  void *image_data;
+
+  /* Content version tracking. Every write use (clEnqWriteBuffer,
+   * clMapBuffer(CL_MAP_WRITE), write_only image, read/write buffers as kernel
+   * args etc) increases the version; read uses do not. At command enqueue
+   * time, the last version across all global mems AND mem_host_ptr
+   * is found and migrated to the destination device's global mem.
+   *
+   * In theory, a simple bool of "valid/invalid could be used;
+   * the only difference is that version saves history
+   * (so we could in future do semi-intelligent memory GC, as in "i see the
+   * buffer on this device hasn't been used for 100 versions, i can free it").
+   */
+  uint64_t version;
+
+  /* Extra pointer for drivers to use for anything
+   *
+   * Currently CUDA uses it to track ALLOC_HOST_PTR allocations.
+   */
+  void *extra_ptr;
+
 } pocl_mem_identifier;
 
+typedef char pixel_t[16];
+
 typedef struct _mem_destructor_callback mem_destructor_callback_t;
 /* represents a memory object destructor callback */
 struct _mem_destructor_callback
@@ -99,8 +126,6 @@ struct _build_program_callback
     void *user_data; /* user supplied data passed to callback function */
 };
 
-// Command Queue datatypes
-
 #define POCL_KERNEL_DIGEST_SIZE 16
 typedef uint8_t pocl_kernel_hash_t[POCL_KERNEL_DIGEST_SIZE];
 
@@ -119,7 +144,6 @@ typedef struct
   int force_generic_wg_func;
   /* If set to 1, disallow "small grid" WG function specialization. */
   int force_large_grid_wg_func;
-  unsigned device_i;
 } _cl_command_run;
 
 // clEnqueueNativeKernel
@@ -127,7 +151,8 @@ typedef struct
 {
   void *args;
   size_t cb_args;
-  void (*user_func)(void *);
+  void **arg_locs;
+  void(CL_CALLBACK *user_func) (void *);
 } _cl_command_native;
 
 // clEnqueueReadBuffer
@@ -153,6 +178,10 @@ typedef struct
 {
   pocl_mem_identifier *src_mem_id;
   pocl_mem_identifier *dst_mem_id;
+  pocl_mem_identifier *src_content_size_mem_id;
+  cl_mem src;
+  cl_mem src_content_size;
+  cl_mem dst;
   size_t src_offset;
   size_t dst_offset;
   size_t size;
@@ -191,6 +220,8 @@ typedef struct
 {
   pocl_mem_identifier *src_mem_id;
   pocl_mem_identifier *dst_mem_id;
+  cl_mem src;
+  cl_mem dst;
   size_t dst_origin[3];
   size_t src_origin[3];
   size_t region[3];
@@ -228,8 +259,10 @@ typedef struct
 typedef struct
 {
   pocl_mem_identifier *src_mem_id;
-  void *__restrict__ dst_host_ptr;
   pocl_mem_identifier *dst_mem_id;
+  void *__restrict__ dst_host_ptr;
+  cl_mem src;
+  cl_mem dst;
   size_t dst_offset;
   size_t origin[3];
   size_t region[3];
@@ -240,8 +273,10 @@ typedef struct
 typedef struct
 {
   pocl_mem_identifier *dst_mem_id;
-  const void *__restrict__ src_host_ptr;
   pocl_mem_identifier *src_mem_id;
+  const void *__restrict__ src_host_ptr;
+  cl_mem src;
+  cl_mem dst;
   size_t src_offset;
   size_t origin[3];
   size_t region[3];
@@ -253,6 +288,8 @@ typedef struct
 {
   pocl_mem_identifier *src_mem_id;
   pocl_mem_identifier *dst_mem_id;
+  cl_mem src;
+  cl_mem dst;
   size_t dst_origin[3];
   size_t src_origin[3];
   size_t region[3];
@@ -261,11 +298,12 @@ typedef struct
 /* clEnqueueFillImage */
 typedef struct
 {
+  pixel_t fill_pixel;
+  cl_uint4 orig_pixel;
+  size_t pixel_size;
   pocl_mem_identifier *mem_id;
   size_t origin[3];
   size_t region[3];
-  void *__restrict__ fill_pixel;
-  size_t pixel_size;
 } _cl_command_fill_image;
 
 /* clEnqueueMarkerWithWaitlist */
@@ -278,19 +316,27 @@ typedef struct
 /* clEnqueueBarrierWithWaitlist */
 typedef _cl_command_marker _cl_command_barrier;
 
+typedef enum pocl_migration_type_e {
+  ENQUEUE_MIGRATE_TYPE_NOP,
+  ENQUEUE_MIGRATE_TYPE_D2H,
+  ENQUEUE_MIGRATE_TYPE_H2D,
+  ENQUEUE_MIGRATE_TYPE_D2D
+} pocl_migration_type_t;
+
 /* clEnqueueMigrateMemObjects */
 typedef struct
 {
-  void *data;
-  size_t num_mem_objects;
-  cl_mem *mem_objects;
-  cl_device_id *source_devices;
+  pocl_migration_type_t type;
+  cl_device_id src_device;
+  pocl_mem_identifier *src_id;
+  pocl_mem_identifier *dst_id;
+  pocl_mem_identifier *mem_id;
 } _cl_command_migrate;
 
 typedef struct
 {
   void* data;
-  void* queue;
+  cl_command_queue queue;
   unsigned  num_svm_pointers;
   void  **svm_pointers;
   void (CL_CALLBACK  *pfn_free_func) ( cl_command_queue queue,
