@@ -31,6 +31,6 @@ POname(clCreateFromGLTexture3D)(cl_context      context,
 CL_API_SUFFIX__VERSION_1_0
 {
   POCL_ABORT_UNIMPLEMENTED("The entire clCreateFromGLTexture3D call");
-  return CL_SUCCESS;
+  return NULL;
 }
 POsym(clCreateFromGLTexture3D)
