@@ -17,8 +17,14 @@ POname(clEnqueueCopyImage)(cl_command_queue      command_queue ,
   cl_device_id device;
   unsigned i;
 
-  POCL_RETURN_ERROR_COND ((src_image == NULL), CL_INVALID_MEM_OBJECT);
-  POCL_RETURN_ERROR_COND ((dst_image == NULL), CL_INVALID_MEM_OBJECT);
+  POCL_RETURN_ERROR_COND ((!IS_CL_OBJECT_VALID (command_queue)),
+                          CL_INVALID_COMMAND_QUEUE);
+
+  POCL_RETURN_ERROR_COND ((!IS_CL_OBJECT_VALID (src_image)),
+                          CL_INVALID_MEM_OBJECT);
+
+  POCL_RETURN_ERROR_COND ((!IS_CL_OBJECT_VALID (dst_image)),
+                          CL_INVALID_MEM_OBJECT);
 
   /* src_image, dst_image: Can be 1D, 2D, 3D image or a 1D, 2D image array
    * objects allowing us to perform the following actions */
@@ -44,8 +50,10 @@ POname(clEnqueueCopyImage)(cl_command_queue      command_queue ,
   if (err != CL_SUCCESS)
     return err;
 
-  cmd->command.copy_image.src_mem_id = &src_image->device_ptrs[device->dev_id];
-  cmd->command.copy_image.dst_mem_id = &dst_image->device_ptrs[device->dev_id];
+  cmd->command.copy_image.src_mem_id = &src_image->device_ptrs[device->global_mem_id];
+  cmd->command.copy_image.src = src_image;
+  cmd->command.copy_image.dst_mem_id = &dst_image->device_ptrs[device->global_mem_id];
+  cmd->command.copy_image.dst = dst_image;
 
   cmd->command.copy_image.src_origin[0] = src_origin[0];
   cmd->command.copy_image.src_origin[1] = src_origin[1];
@@ -57,11 +65,6 @@ POname(clEnqueueCopyImage)(cl_command_queue      command_queue ,
   cmd->command.copy_image.region[1] = region[1];
   cmd->command.copy_image.region[2] = region[2];
 
-  POname (clRetainMemObject) (src_image);
-  src_image->owning_device = device;
-  POname (clRetainMemObject) (dst_image);
-  dst_image->owning_device = device;
-
   pocl_command_enqueue (command_queue, cmd);
 
   return CL_SUCCESS;
