@@ -17,9 +17,11 @@ POname(clEnqueueCopyImageToBuffer)(cl_command_queue  command_queue ,
   /* pass dst_origin through in a format pocl_rect_copy understands */
   const size_t dst_origin[3] = { dst_offset, 0, 0};
 
-  POCL_RETURN_ERROR_COND ((command_queue == NULL), CL_INVALID_COMMAND_QUEUE);
+  POCL_RETURN_ERROR_COND ((!IS_CL_OBJECT_VALID (command_queue)),
+                          CL_INVALID_COMMAND_QUEUE);
 
-  POCL_RETURN_ERROR_COND ((src_image == NULL), CL_INVALID_MEM_OBJECT);
+  POCL_RETURN_ERROR_COND ((!IS_CL_OBJECT_VALID (src_image)),
+                          CL_INVALID_MEM_OBJECT);
 
   if (IS_IMAGE1D_BUFFER (src_image))
     {
@@ -60,9 +62,11 @@ POname(clEnqueueCopyImageToBuffer)(cl_command_queue  command_queue ,
 
   cl_device_id dev = command_queue->device;
 
-  cmd->command.read_image.src_mem_id = &src_image->device_ptrs[dev->dev_id];
+  cmd->command.read_image.src_mem_id = &src_image->device_ptrs[dev->global_mem_id];
+  cmd->command.read_image.src = src_image;
   cmd->command.read_image.dst_host_ptr = NULL;
-  cmd->command.read_image.dst_mem_id = &dst_buffer->device_ptrs[dev->dev_id];
+  cmd->command.read_image.dst = dst_buffer;
+  cmd->command.read_image.dst_mem_id = &dst_buffer->device_ptrs[dev->global_mem_id];
 
   cmd->command.read_image.origin[0] = src_origin[0];
   cmd->command.read_image.origin[1] = src_origin[1];
@@ -75,11 +79,6 @@ POname(clEnqueueCopyImageToBuffer)(cl_command_queue  command_queue ,
   cmd->command.read_image.dst_slice_pitch = 0; // src_image->image_slice_pitch;
   cmd->command.read_image.dst_offset = dst_offset;
 
-  POname (clRetainMemObject) (src_image);
-  src_image->owning_device = dev;
-  POname (clRetainMemObject) (dst_buffer);
-  dst_buffer->owning_device = dev;
-
   pocl_command_enqueue (command_queue, cmd);
 
   return CL_SUCCESS;
