@@ -33,7 +33,8 @@ POname(clGetCommandQueueInfo)(cl_command_queue      command_queue ,
                       void *                param_value ,
                       size_t *              param_value_size_ret) CL_API_SUFFIX__VERSION_1_0
 {
-  POCL_RETURN_ERROR_COND((command_queue == NULL), CL_INVALID_COMMAND_QUEUE);
+  POCL_RETURN_ERROR_COND ((!IS_CL_OBJECT_VALID (command_queue)),
+                          CL_INVALID_COMMAND_QUEUE);
 
   switch (param_name)
   {
