@@ -59,7 +59,7 @@ POname(clGetDeviceInfo)(cl_device_id   device,
                 void *         param_value,
                 size_t *       param_value_size_ret) CL_API_SUFFIX__VERSION_1_0
 {
-  POCL_RETURN_ERROR_COND ((device == NULL), CL_INVALID_DEVICE);
+  POCL_RETURN_ERROR_COND ((!IS_CL_OBJECT_VALID (device)), CL_INVALID_DEVICE);
 
   switch (param_name)
   {
@@ -186,15 +186,10 @@ POname(clGetDeviceInfo)(cl_device_id   device,
   case CL_DEVICE_COMPILER_AVAILABLE                :
     POCL_RETURN_GETINFO(cl_bool, device->compiler_available);
   case CL_DEVICE_LINKER_AVAILABLE                  :
-    /* TODO currently we return the same availability as the compiler,
-     * since if the compiler is available the linker MUST be available
-     * too. The only case where the linker and compiler availability can
-     * be different is when the linker is available and the compiler is not,
-     * which is not the case in pocl currently */
-    POCL_RETURN_GETINFO(cl_bool, device->compiler_available);
+    POCL_RETURN_GETINFO (cl_bool, device->linker_available);
   case CL_DEVICE_EXECUTION_CAPABILITIES            :
     POCL_RETURN_GETINFO(cl_device_exec_capabilities, device->execution_capabilities);
-   
+
   case CL_DEVICE_NAME:
     POCL_RETURN_GETINFO_STR(device->long_name);
    
@@ -248,7 +243,10 @@ POname(clGetDeviceInfo)(cl_device_id   device,
   case CL_DEVICE_OPENCL_C_VERSION                  :
     POCL_RETURN_GETINFO_STR (HOST_CL_VERSION);
   case CL_DEVICE_BUILT_IN_KERNELS                  :
-    POCL_RETURN_GETINFO_STR("");
+    if (device->builtin_kernel_list)
+      POCL_RETURN_GETINFO_STR (device->builtin_kernel_list);
+    else
+      POCL_RETURN_GETINFO_STR ("");
 
   case CL_DEVICE_PARENT_DEVICE                     :
     POCL_RETURN_GETINFO(cl_device_id, device->parent_device);
@@ -287,6 +285,10 @@ POname(clGetDeviceInfo)(cl_device_id   device,
 
   case CL_DEVICE_SVM_CAPABILITIES:
     POCL_RETURN_GETINFO(cl_device_svm_capabilities, device->svm_caps);
+  case CL_DEVICE_ATOMIC_MEMORY_CAPABILITIES:
+    POCL_RETURN_GETINFO(cl_device_atomic_capabilities, device->atomic_memory_capabilities);
+  case CL_DEVICE_ATOMIC_FENCE_CAPABILITIES:
+    POCL_RETURN_GETINFO(cl_device_atomic_capabilities, device->atomic_fence_capabilities);
   case CL_DEVICE_MAX_ON_DEVICE_EVENTS:
     POCL_RETURN_GETINFO(cl_uint, device->max_events);
   case CL_DEVICE_MAX_ON_DEVICE_QUEUES:
@@ -322,8 +324,17 @@ POname(clGetDeviceInfo)(cl_device_id   device,
   case CL_DEVICE_MAX_GLOBAL_VARIABLE_SIZE:
     POCL_RETURN_GETINFO(size_t, device->global_var_max_size);
   case CL_DEVICE_IL_VERSION:
-    POCL_RETURN_GETINFO_STR (device->spirv_version);
+    if (device->spirv_version)
+      POCL_RETURN_GETINFO_STR (device->spirv_version);
+    else
+      POCL_RETURN_GETINFO_STR ("");
   }
+
+  if(device->ops->get_device_info_ext != NULL) {
+    return device->ops->get_device_info_ext(device, param_name, param_value_size,
+                                            param_value, param_value_size_ret);
+  }
+
   return CL_INVALID_VALUE;
 }
 POsym(clGetDeviceInfo)
