@@ -39,6 +39,7 @@ CL_EXT_SUFFIX__VERSION_1_2
       return NULL;
     }
 
+  assert (pocl_platform);
   if (platform != pocl_platform)
     {
       POCL_MSG_PRINT_INFO ("Requested Function Address not "
@@ -50,6 +51,10 @@ CL_EXT_SUFFIX__VERSION_1_2
   if (strcmp (func_name, "clIcdGetPlatformIDsKHR") == 0)
     return (void *)&POname(clIcdGetPlatformIDsKHR);
 #endif
+
+  if (strcmp (func_name, "clSetContentSizeBufferPoCL") == 0)
+    return (void *)&POname (clSetContentSizeBufferPoCL);
+
   if (strcmp (func_name, "clGetPlatformInfo") == 0)
     return (void *)&POname(clGetPlatformInfo);
 
