@@ -33,7 +33,7 @@ POname(clGetKernelInfo)(cl_kernel      kernel ,
                 void *         param_value ,
                 size_t *       param_value_size_ret) CL_API_SUFFIX__VERSION_1_0
 {
-  POCL_RETURN_ERROR_COND((kernel == NULL), CL_INVALID_KERNEL);
+  POCL_RETURN_ERROR_COND ((!IS_CL_OBJECT_VALID (kernel)), CL_INVALID_KERNEL);
 
   switch (param_name) {
   case CL_KERNEL_FUNCTION_NAME:
