@@ -36,7 +36,7 @@ POname(clGetKernelWorkGroupInfo)
   CL_API_SUFFIX__VERSION_1_0
 {
 
-  POCL_RETURN_ERROR_COND ((kernel == NULL), CL_INVALID_KERNEL);
+  POCL_RETURN_ERROR_COND ((!IS_CL_OBJECT_VALID (kernel)), CL_INVALID_KERNEL);
 
   /* Check that kernel is associated with device, or that there is no
      risk of confusion. */
