@@ -24,18 +24,27 @@
 #include "pocl_cl.h"
 #include "pocl_util.h"
 
+extern unsigned long kernel_c;
+
 CL_API_ENTRY cl_int CL_API_CALL
 POname(clReleaseKernel)(cl_kernel kernel) CL_API_SUFFIX__VERSION_1_0
 {
   int new_refcount;
   unsigned i;
 
-  POCL_RETURN_ERROR_COND((kernel == NULL), CL_INVALID_KERNEL);
+  POCL_RETURN_ERROR_COND ((!IS_CL_OBJECT_VALID (kernel)), CL_INVALID_KERNEL);
+
   POCL_RELEASE_OBJECT (kernel, new_refcount);
   POCL_MSG_PRINT_REFCOUNTS ("Release kernel %p  %d\n", kernel, new_refcount);
 
   if (new_refcount == 0)
     {
+      VG_REFC_ZERO (kernel);
+
+      POCL_ATOMIC_DEC (kernel_c);
+
+      TP_FREE_KERNEL (kernel->context->id, kernel->id, kernel->name);
+
       POCL_MSG_PRINT_REFCOUNTS ("Free kernel %p\n", kernel);
       cl_program program = kernel->program;
       assert (program != NULL);
@@ -44,14 +53,28 @@ POname(clReleaseKernel)(cl_kernel kernel) CL_API_SUFFIX__VERSION_1_0
       POCL_LOCK_OBJ (program);
       LL_DELETE (program->kernels, kernel);
 
-      if (kernel->dyn_arguments)
-        for (i = 0; i < (kernel->meta->num_args); i++)
-          {
-            pocl_aligned_free (kernel->dyn_arguments[i].value);
-          }
+      for (i = 0; i < program->num_devices; ++i)
+        {
+          cl_device_id device = program->devices[i];
+          if (device->ops->free_kernel)
+            device->ops->free_kernel (device, program, kernel, i);
+        }
 
+      if (kernel->meta->total_argument_storage_size)
+        {
+          POCL_MEM_FREE (kernel->dyn_argument_storage);
+          POCL_MEM_FREE (kernel->dyn_argument_offsets);
+        }
+      else
+        {
+          for (i = 0; i < (kernel->meta->num_args); i++)
+            {
+              pocl_aligned_free (kernel->dyn_arguments[i].value);
+            }
+        }
       kernel->name = NULL;
       kernel->meta = NULL;
+      POCL_MEM_FREE (kernel->data);
       POCL_MEM_FREE (kernel->dyn_arguments);
       POCL_DESTROY_OBJECT (kernel);
       POCL_MEM_FREE (kernel);
@@ -59,7 +82,11 @@ POname(clReleaseKernel)(cl_kernel kernel) CL_API_SUFFIX__VERSION_1_0
 
       POname(clReleaseProgram) (program);
     }
-  
+  else
+    {
+      VG_REFC_NONZERO (kernel);
+    }
+
   return CL_SUCCESS;
 }
 POsym(clReleaseKernel)
