@@ -26,7 +26,7 @@
 CL_API_ENTRY cl_int CL_API_CALL
 POname(clRetainContext)(cl_context context) CL_API_SUFFIX__VERSION_1_0
 {
-  POCL_RETURN_ERROR_COND((context == NULL), CL_INVALID_CONTEXT);
+  POCL_RETURN_ERROR_COND ((!IS_CL_OBJECT_VALID (context)), CL_INVALID_CONTEXT);
   int refc;
   POCL_RETAIN_OBJECT_REFCOUNT (context, refc);
   POCL_MSG_PRINT_REFCOUNTS ("Retain Context %p  : %d\n", context, refc);
