@@ -28,9 +28,6 @@
 CL_API_ENTRY cl_int CL_API_CALL POname (clUnloadCompiler) (void)
     CL_API_SUFFIX__VERSION_1_1
 {
-#ifdef OCS_AVAILABLE
-  pocl_llvm_release ();
-#endif
   pocl_check_uninit_devices ();
   return CL_SUCCESS;
 }
