@@ -60,61 +60,57 @@
 extern "C" {
 #endif
 
-int llvm_codegen (char *output, unsigned device_i, cl_kernel kernel,
-                  cl_device_id device, _cl_command_node *command,
-                  int specialize);
+POCL_EXPORT
+void pocl_fill_dev_image_t (dev_image_t *di, struct pocl_argument *parg,
+                            cl_device_id device);
 
-void fill_dev_image_t (dev_image_t *di, struct pocl_argument *parg,
-                       cl_device_id device);
+POCL_EXPORT
+void pocl_fill_dev_sampler_t (dev_sampler_t *ds, struct pocl_argument *parg);
 
-void fill_dev_sampler_t (dev_sampler_t *ds, struct pocl_argument *parg);
-
-void pocl_copy_mem_object (cl_device_id dest_dev, cl_mem dest,
-                           size_t dest_offset,
-                           cl_device_id source_dev, cl_mem source,
-                           size_t source_offset, size_t cb);
-
-void pocl_migrate_mem_objects (_cl_command_node *node);
-
-void pocl_scheduler (_cl_command_node * volatile * ready_list,
-                     pthread_mutex_t *lock_ptr);
-
-void pocl_exec_command (_cl_command_node * volatile node);
+POCL_EXPORT
+void pocl_exec_command (_cl_command_node *node);
 
+POCL_EXPORT
 void pocl_ndrange_node_cleanup(_cl_command_node *node);
-void pocl_mem_objs_cleanup (cl_event event);
 
+POCL_EXPORT
 void pocl_broadcast (cl_event event);
 
+POCL_EXPORT
 void pocl_init_dlhandle_cache ();
 
+POCL_EXPORT
 char *pocl_check_kernel_disk_cache (_cl_command_node *cmd, int specialized);
 
 size_t pocl_cmd_max_grid_dim_width (_cl_command_run *cmd);
 
+POCL_EXPORT
 void pocl_check_kernel_dlhandle_cache (_cl_command_node *command,
                                        unsigned initial_refcount,
                                        int specialize);
 
+POCL_EXPORT
 void pocl_release_dlhandle_cache (_cl_command_node *cmd);
 
 void pocl_setup_device_for_system_memory(cl_device_id device);
 
 void pocl_reinit_system_memory();
 
+POCL_EXPORT
 void pocl_set_buffer_image_limits(cl_device_id device);
 
+POCL_EXPORT
 void* pocl_aligned_malloc_global_mem(cl_device_id device, size_t align, size_t size);
 
+POCL_EXPORT
 void pocl_free_global_mem(cl_device_id device, void *ptr, size_t size);
 
 void pocl_print_system_memory_stats();
 
-void pocl_calculate_kernel_hash (cl_program program, unsigned kernel_i,
-                                 unsigned device_i);
-
+POCL_EXPORT
 void pocl_init_default_device_infos (cl_device_id dev);
 
+
 #ifdef __cplusplus
 }
 #endif
