@@ -26,6 +26,7 @@
 #include "config.h"
 
 #include "common.h"
+#include "common_driver.h"
 #include "devices.h"
 #include "pocl.h"
 #include "pocl-cuda.h"
@@ -41,6 +42,7 @@
 #include <string.h>
 
 #include <cuda.h>
+#include <cuda_runtime.h>
 
 typedef struct pocl_cuda_device_data_s
 {
@@ -81,6 +83,7 @@ typedef struct pocl_cuda_event_data_s
   CUevent end;
   volatile int events_ready;
   cl_int *ext_event_flag;
+  pthread_cond_t event_cond;
   volatile unsigned num_ext_events;
 } pocl_cuda_event_data_t;
 
@@ -126,6 +129,67 @@ pocl_cuda_error (CUresult result, unsigned line, const char *func,
 #define CUDA_CHECK_ERROR(result, api)                                         \
   pocl_cuda_error (result, __LINE__, __FUNCTION__, #result, api)
 
+cl_int pocl_cuda_handle_cl_nv_device_attribute_query(cl_device_id   device,
+                                                     cl_device_info param_name,
+                                                     size_t         param_value_size,
+                                                     void *         param_value,
+                                                     size_t *       param_value_size_ret)
+{
+  CUdevice cudaDev = ((pocl_cuda_device_data_t *)device->data)->device;
+  unsigned int value;
+  CUresult res;
+
+  switch(param_name) {
+    case CL_DEVICE_COMPUTE_CAPABILITY_MAJOR_NV:
+      res = cuDeviceGetAttribute(&value, CU_DEVICE_ATTRIBUTE_COMPUTE_CAPABILITY_MAJOR, cudaDev);
+      CUDA_CHECK(res, "cuDeviceGetAttribute");
+      POCL_RETURN_GETINFO(cl_uint, value);
+    case CL_DEVICE_COMPUTE_CAPABILITY_MINOR_NV:
+      res = cuDeviceGetAttribute(&value, CU_DEVICE_ATTRIBUTE_COMPUTE_CAPABILITY_MINOR, cudaDev);
+      CUDA_CHECK(res, "cuDeviceGetAttribute");
+      POCL_RETURN_GETINFO(cl_uint, value);
+    case CL_DEVICE_REGISTERS_PER_BLOCK_NV:
+      res = cuDeviceGetAttribute(&value, CU_DEVICE_ATTRIBUTE_MAX_REGISTERS_PER_BLOCK, cudaDev);
+      CUDA_CHECK(res, "cuDeviceGetAttribute");
+      POCL_RETURN_GETINFO(cl_uint, value);
+    case CL_DEVICE_WARP_SIZE_NV:
+      res = cuDeviceGetAttribute(&value, CU_DEVICE_ATTRIBUTE_WARP_SIZE, cudaDev);
+      CUDA_CHECK(res, "cuDeviceGetAttribute");
+      POCL_RETURN_GETINFO(cl_uint, value);
+    case CL_DEVICE_GPU_OVERLAP_NV:
+      res = cuDeviceGetAttribute(&value, CU_DEVICE_ATTRIBUTE_GPU_OVERLAP, cudaDev);
+      CUDA_CHECK(res, "cuDeviceGetAttribute");
+      POCL_RETURN_GETINFO(cl_bool, value);
+    case CL_DEVICE_KERNEL_EXEC_TIMEOUT_NV:
+      res = cuDeviceGetAttribute(&value, CU_DEVICE_ATTRIBUTE_KERNEL_EXEC_TIMEOUT, cudaDev);
+      CUDA_CHECK(res, "cuDeviceGetAttribute");
+      POCL_RETURN_GETINFO(cl_bool, value);
+    case CL_DEVICE_INTEGRATED_MEMORY_NV:
+      res = cuDeviceGetAttribute(&value, CU_DEVICE_ATTRIBUTE_INTEGRATED, cudaDev);
+      CUDA_CHECK(res, "cuDeviceGetAttribute");
+      POCL_RETURN_GETINFO(cl_bool, value);
+    case CL_DEVICE_ATTRIBUTE_ASYNC_ENGINE_COUNT_NV:
+      res = cuDeviceGetAttribute(&value, CU_DEVICE_ATTRIBUTE_ASYNC_ENGINE_COUNT, cudaDev);
+      CUDA_CHECK(res, "cuDeviceGetAttribute");
+      POCL_RETURN_GETINFO(cl_uint, value);
+    case CL_DEVICE_PCI_BUS_ID_NV:
+      res = cuDeviceGetAttribute(&value, CU_DEVICE_ATTRIBUTE_PCI_BUS_ID, cudaDev);
+      CUDA_CHECK(res, "cuDeviceGetAttribute");
+      POCL_RETURN_GETINFO(cl_uint, value);
+    case CL_DEVICE_PCI_SLOT_ID_NV:
+      res = cuDeviceGetAttribute(&value, CU_DEVICE_ATTRIBUTE_PCI_DEVICE_ID, cudaDev);
+      CUDA_CHECK(res, "cuDeviceGetAttribute");
+      POCL_RETURN_GETINFO(cl_uint, value);
+    case CL_DEVICE_PCI_DOMAIN_ID_NV:
+      res = cuDeviceGetAttribute(&value, CU_DEVICE_ATTRIBUTE_PCI_DOMAIN_ID, cudaDev);
+      CUDA_CHECK(res, "cuDeviceGetAttribute");
+      POCL_RETURN_GETINFO(cl_uint, value);
+    default:
+      return CL_INVALID_VALUE;
+  }
+
+}
+
 void
 pocl_cuda_init_device_ops (struct pocl_device_ops *ops)
 {
@@ -137,9 +201,10 @@ pocl_cuda_init_device_ops (struct pocl_device_ops *ops)
   ops->init = pocl_cuda_init;
   ops->init_queue = pocl_cuda_init_queue;
   ops->free_queue = pocl_cuda_free_queue;
+
   ops->alloc_mem_obj = pocl_cuda_alloc_mem_obj;
   ops->free = pocl_cuda_free;
-  ops->compile_kernel = pocl_cuda_compile_kernel;
+
   ops->submit = pocl_cuda_submit;
   ops->notify = pocl_cuda_notify;
   ops->broadcast = pocl_broadcast;
@@ -150,14 +215,31 @@ pocl_cuda_init_device_ops (struct pocl_device_ops *ops)
   ops->flush = pocl_cuda_flush;
   ops->init_build = pocl_cuda_init_build;
   // TODO
-  ops->map_mem = pocl_cuda_map_mem;
+  ops->notify_event_finished = pocl_cuda_notify_event_finished;
 
+  ops->get_device_info_ext = pocl_cuda_handle_cl_nv_device_attribute_query;
+  ops->build_source = pocl_driver_build_source;
+  ops->link_program = pocl_driver_link_program;
+  ops->build_binary = pocl_driver_build_binary;
+  ops->free_program = pocl_driver_free_program;
+  ops->setup_metadata = pocl_driver_setup_metadata;
+  ops->supports_binary = pocl_driver_supports_binary;
+  ops->build_poclbinary = pocl_driver_build_poclbinary;
+  ops->compile_kernel = pocl_cuda_compile_kernel;
+
+  // TODO
+  ops->get_mapping_ptr = pocl_driver_get_mapping_ptr;
+  ops->free_mapping_ptr = pocl_driver_free_mapping_ptr;
+
+  ops->can_migrate_d2d = NULL;
+  ops->migrate_d2d = NULL;
   ops->read = NULL;
   ops->read_rect = NULL;
   ops->write = NULL;
   ops->write_rect = NULL;
   ops->copy = NULL;
   ops->copy_rect = NULL;
+  ops->map_mem = NULL;
   ops->unmap_mem = NULL;
   ops->run = NULL;
 }
@@ -292,7 +374,7 @@ pocl_cuda_init (unsigned j, cl_device_id dev, const char *parameters)
       if (ret != CL_INVALID_DEVICE)
         {
           POCL_MSG_ERR ("[CUDA] failed to find libdevice library\n");
-          dev->compiler_available = 0;
+          dev->compiler_available = dev->linker_available = 0;
         }
     }
 
@@ -334,7 +416,7 @@ pocl_cuda_init (unsigned j, cl_device_id dev, const char *parameters)
 }
 
 cl_int
-pocl_cuda_init_queue (cl_command_queue queue)
+pocl_cuda_init_queue (cl_device_id device, cl_command_queue queue)
 {
   cuCtxSetCurrent (((pocl_cuda_device_data_t *)queue->device->data)->context);
 
@@ -376,8 +458,8 @@ pocl_cuda_init_queue (cl_command_queue queue)
   return CL_SUCCESS;
 }
 
-void
-pocl_cuda_free_queue (cl_command_queue queue)
+int
+pocl_cuda_free_queue (cl_device_id device, cl_command_queue queue)
 {
   pocl_cuda_queue_data_t *queue_data = (pocl_cuda_queue_data_t *)queue->data;
 
@@ -398,6 +480,7 @@ pocl_cuda_free_queue (cl_command_queue queue)
       pthread_join (queue_data->submit_thread, NULL);
       pthread_join (queue_data->finalize_thread, NULL);
     }
+  return CL_SUCCESS;
 }
 
 char *
@@ -453,20 +536,21 @@ pocl_cuda_uninit (unsigned j, cl_device_id device)
 }
 
 cl_int
-pocl_cuda_alloc_mem_obj (cl_device_id device, cl_mem mem_obj, void *host_ptr)
+pocl_cuda_alloc_mem_obj (cl_device_id device, cl_mem mem, void *host_ptr)
 {
   cuCtxSetCurrent (((pocl_cuda_device_data_t *)device->data)->context);
+  pocl_mem_identifier *p = &mem->device_ptrs[device->global_mem_id];
+  int err = CL_MEM_OBJECT_ALLOCATION_FAILURE;
 
   CUresult result;
   void *b = NULL;
 
-  /* If memory for this global memory is not yet allocated -> do it */
-  if (mem_obj->device_ptrs[device->global_mem_id].mem_ptr == NULL)
-    {
-      cl_mem_flags flags = mem_obj->flags;
+  p->extra_ptr = NULL;
+  p->version = 0;
+  cl_mem_flags flags = mem->flags;
 
-      if (flags & CL_MEM_USE_HOST_PTR)
-        {
+  if (flags & CL_MEM_USE_HOST_PTR)
+    {
 #if defined __arm__
           /* cuMemHostRegister is not supported on ARM.
            * Allocate device memory and perform explicit copies
@@ -474,86 +558,98 @@ pocl_cuda_alloc_mem_obj (cl_device_id device, cl_mem mem_obj, void *host_ptr)
           result = cuMemAlloc ((CUdeviceptr *)&b, mem_obj->size);
           CUDA_CHECK (result, "cuMemAlloc");
 #else
-          result = cuMemHostRegister (host_ptr, mem_obj->size,
-                                      CU_MEMHOSTREGISTER_DEVICEMAP);
-          if (result != CUDA_SUCCESS
-              && result != CUDA_ERROR_HOST_MEMORY_ALREADY_REGISTERED)
-            CUDA_CHECK (result, "cuMemHostRegister");
-          result = cuMemHostGetDevicePointer ((CUdeviceptr *)&b, host_ptr, 0);
-          CUDA_CHECK (result, "cuMemHostGetDevicePointer");
+      POCL_RETURN_ERROR_ON ((pocl_alloc_or_retain_mem_host_ptr (mem) != 0),
+                            CL_OUT_OF_HOST_MEMORY,
+                            "Cannot allocate backing memory!\n");
+
+      result = cuMemHostRegister (mem->mem_host_ptr, mem->size,
+                                  CU_MEMHOSTREGISTER_DEVICEMAP);
+      if (result != CUDA_SUCCESS
+          && result != CUDA_ERROR_HOST_MEMORY_ALREADY_REGISTERED)
+        CUDA_CHECK (result, "cuMemHostRegister");
+      result = cuMemHostGetDevicePointer ((CUdeviceptr *)&b, mem->mem_host_ptr,
+                                          0);
+      CUDA_CHECK (result, "cuMemHostGetDevicePointer");
+
+      /* TODO can we assume cuMemHostRegister copies
+       * the content of host memory to the device ? for now, lets not */
+      p->version = 0;
 #endif
-        }
-      else if (flags & CL_MEM_ALLOC_HOST_PTR)
-        {
-          result = cuMemHostAlloc (&mem_obj->mem_host_ptr, mem_obj->size,
-                                   CU_MEMHOSTREGISTER_DEVICEMAP);
-          CUDA_CHECK (result, "cuMemHostAlloc");
-          result = cuMemHostGetDevicePointer ((CUdeviceptr *)&b,
-                                              mem_obj->mem_host_ptr, 0);
-          CUDA_CHECK (result, "cuMemHostGetDevicePointer");
-        }
-      else
-        {
-          result = cuMemAlloc ((CUdeviceptr *)&b, mem_obj->size);
-          if (result != CUDA_SUCCESS)
-            {
-              const char *err;
-              cuGetErrorName (result, &err);
-              POCL_MSG_PRINT2 (CUDA, __FUNCTION__, __LINE__,
-                               "-> Failed to allocate memory: %s\n", err);
-              return CL_MEM_OBJECT_ALLOCATION_FAILURE;
-            }
-        }
+    }
+  /* preallocate host visible memory */
+  else if ((flags & CL_MEM_ALLOC_HOST_PTR) && (mem->mem_host_ptr == NULL))
+    {
+      result = cuMemHostAlloc (&p->extra_ptr, mem->size,
+                               CU_MEMHOSTREGISTER_DEVICEMAP);
+      CUDA_CHECK (result, "cuMemHostAlloc");
+      result = cuMemHostGetDevicePointer ((CUdeviceptr *)&b, p->extra_ptr, 0);
+      CUDA_CHECK (result, "cuMemHostGetDevicePointer");
+      mem->mem_host_ptr = p->extra_ptr;
+      mem->mem_host_ptr_refcount = 1;
+      mem->mem_host_ptr_version = 0;
 
       if (flags & CL_MEM_COPY_HOST_PTR)
         {
-          result = cuMemcpyHtoD ((CUdeviceptr)b, host_ptr, mem_obj->size);
+          result = cuMemcpyHtoD ((CUdeviceptr)b, host_ptr, mem->size);
           CUDA_CHECK (result, "cuMemcpyHtoD");
 
           result = cuStreamSynchronize (0);
           CUDA_CHECK (result, "cuStreamSynchronize");
+
+          mem->mem_host_ptr_version = 1;
+          mem->latest_version = 1;
+          p->version = 1;
         }
-
-      mem_obj->device_ptrs[device->global_mem_id].mem_ptr = b;
-      mem_obj->device_ptrs[device->global_mem_id].global_mem_id
-          = device->global_mem_id;
     }
+  else
+    {
+      result = cuMemAlloc ((CUdeviceptr *)&b, mem->size);
+      if (result != CUDA_SUCCESS)
+        {
+          const char *err;
+          cuGetErrorName (result, &err);
+          POCL_MSG_PRINT2 (CUDA, __FUNCTION__, __LINE__,
+                           "-> Failed to allocate memory: %s\n", err);
+          return CL_MEM_OBJECT_ALLOCATION_FAILURE;
+        }
+    }
+  p->mem_ptr = b;
+  err = CL_SUCCESS;
 
-  /* Copy allocated global mem info to devices own slot */
-  mem_obj->device_ptrs[device->dev_id]
-      = mem_obj->device_ptrs[device->global_mem_id];
-
-  return CL_SUCCESS;
+  return err;
 }
 
 void
 pocl_cuda_free (cl_device_id device, cl_mem mem_obj)
 {
   cuCtxSetCurrent (((pocl_cuda_device_data_t *)device->data)->context);
+  pocl_mem_identifier *p = &mem_obj->device_ptrs[device->global_mem_id];
 
-  if (mem_obj->flags & CL_MEM_ALLOC_HOST_PTR)
-    {
-      cuMemFreeHost (mem_obj->mem_host_ptr);
-      mem_obj->mem_host_ptr = NULL;
-    }
-  else if (mem_obj->flags & CL_MEM_USE_HOST_PTR)
+  if (mem_obj->flags & CL_MEM_USE_HOST_PTR)
     {
+#if defined __arm__
+      cuMemFree ((CUdeviceptr)p->mem_ptr);
+#else
+      assert (p->extra_ptr == NULL);
       cuMemHostUnregister (mem_obj->mem_host_ptr);
+#endif
+    }
+  else if (p->extra_ptr)
+    {
       mem_obj->mem_host_ptr = NULL;
+      mem_obj->mem_host_ptr_refcount = 0;
+      mem_obj->mem_host_ptr_version = 0;
+      cuMemFreeHost (p->extra_ptr);
+      p->extra_ptr = NULL;
     }
   else
     {
-      void *ptr = mem_obj->device_ptrs[device->dev_id].mem_ptr;
-      cuMemFree ((CUdeviceptr)ptr);
+      assert (p->extra_ptr == NULL);
+      assert (p->mem_ptr != NULL);
+      cuMemFree ((CUdeviceptr)p->mem_ptr);
     }
-}
-
-void
-pocl_cuda_free_ptr (cl_device_id device, void *mem_ptr)
-{
-  cuCtxSetCurrent (((pocl_cuda_device_data_t *)device->data)->context);
-
-  cuMemFreeHost (mem_ptr);
+  p->mem_ptr = NULL;
+  p->version = 0;
 }
 
 void
@@ -749,32 +845,20 @@ pocl_cuda_submit_copy_rect (CUstream stream,
   CUDA_CHECK (result, "cuMemcpy3DAsync");
 }
 
-cl_int
-pocl_cuda_map_mem (void *data,
-                    pocl_mem_identifier *src_mem_id,
-                    cl_mem src_buf,
-                    mem_mapping_t *map)
-{
-  void *host_ptr = map->host_ptr;
-
-  assert (host_ptr == NULL);
-
-  map->host_ptr = (char *)malloc (map->size);
-  return CL_SUCCESS;
-}
-
 void
-pocl_cuda_submit_map_mem (CUstream stream, pocl_mem_identifier *mem,
+pocl_cuda_submit_map_mem (CUstream stream, cl_mem mem,
+                          pocl_mem_identifier *p,
                           size_t offset, size_t size, void *host_ptr)
 {
   assert (host_ptr != NULL);
 
-  /* TODO: Map instead of copy? */
-  /* TODO: don't copy if mapped as CL_MAP_WRITE_INVALIDATE_REGION */
+  if ((mem->flags & CL_MEM_USE_HOST_PTR)
+      || (p->extra_ptr))
+    return;
 
   POCL_MSG_PRINT_CUDA ("cuMemcpyDtoHAsync %p / %zu B \n", host_ptr, size);
 
-  void *buf_ptr = mem->mem_ptr;
+  void *buf_ptr = p->mem_ptr;
 
   CUresult result = cuMemcpyDtoHAsync (
       host_ptr, (CUdeviceptr) (buf_ptr + offset), size, stream);
@@ -783,11 +867,15 @@ pocl_cuda_submit_map_mem (CUstream stream, pocl_mem_identifier *mem,
 
 void *
 pocl_cuda_submit_unmap_mem (CUstream stream, pocl_mem_identifier *dst_mem_id,
-                            size_t offset, size_t size, void *host_ptr)
+                            size_t offset, size_t size, void *host_ptr,
+                            cl_map_flags map_flags)
 {
+  /* Only copy back if mapped for writing */
+  if (map_flags == CL_MAP_READ)
+    return NULL;
+
   if (host_ptr)
     {
-      /* TODO: Only copy back if mapped for writing */
       CUresult result = cuMemcpyHtoDAsync (
           (CUdeviceptr) (dst_mem_id->mem_ptr + offset), host_ptr, size, stream);
       CUDA_CHECK (result, "cuMemcpyHtoDAsync");
@@ -891,7 +979,7 @@ void
 pocl_cuda_compile_kernel (_cl_command_node *cmd, cl_kernel kernel,
                           cl_device_id device, int specialize)
 {
-  load_or_generate_kernel (kernel, device, 0, cmd->command.run.device_i, cmd,
+  load_or_generate_kernel (kernel, device, 0, cmd->device_i, cmd,
                            specialize);
 }
 
@@ -911,7 +999,7 @@ pocl_cuda_submit_kernel (CUstream stream, _cl_command_node *cmd,
 
   /* Get kernel function */
   pocl_cuda_kernel_data_t *kdata = load_or_generate_kernel (
-      kernel, device, has_offsets, run.device_i, cmd, 1);
+      kernel, device, has_offsets, cmd->device_i, cmd, 1);
   CUmodule module = has_offsets ? kdata->module_offsets : kdata->module;
   CUfunction function = has_offsets ? kdata->kernel_offsets : kdata->kernel;
 
@@ -960,11 +1048,12 @@ pocl_cuda_submit_kernel (CUstream stream, _cl_command_node *cmd,
                      == CL_KERNEL_ARG_ADDRESS_CONSTANT)
               {
                 assert (constant_mem_base);
+                assert (arguments[i].is_svm == 0);
 
                 /* Get device pointer */
                 cl_mem mem = *(void **)arguments[i].value;
                 CUdeviceptr src
-                    = (CUdeviceptr)mem->device_ptrs[device->dev_id].mem_ptr
+                    = (CUdeviceptr)mem->device_ptrs[device->global_mem_id].mem_ptr
                       + arguments[i].offset;
 
                 size_t align = kdata->alignments[i];
@@ -986,10 +1075,11 @@ pocl_cuda_submit_kernel (CUstream stream, _cl_command_node *cmd,
               }
             else
               {
+                assert (arguments[i].is_svm == 0);
                 if (arguments[i].value)
                   {
                     cl_mem mem = *(void **)arguments[i].value;
-                    params[i] = &mem->device_ptrs[device->dev_id].mem_ptr
+                    params[i] = &mem->device_ptrs[device->global_mem_id].mem_ptr
                                 + arguments[i].offset;
 
 #if defined __arm__
@@ -1209,35 +1299,12 @@ pocl_cuda_submit_node (_cl_command_node *node, cl_command_queue cq, int locked)
           cmd->copy_rect.dst_slice_pitch);
         break;
       }
-    case CL_COMMAND_MIGRATE_MEM_OBJECTS:
-      {
-        size_t i;
-        for (i = 0; i < cmd->migrate.num_mem_objects; i++)
-          {
-//            cl_device_id src_dev = cmd->migrate.source_devices[i];
-//            cl_device_id dst_dev = cq->device;
-            cl_mem buf = cmd->migrate.mem_objects[i];
-//            if (!src_dev)
-//              src_dev = dst_dev;
-            // TODO
-            void *ptr = buf->device_ptrs[dev->dev_id].mem_ptr;
-            pocl_cuda_submit_copy (stream, ptr, 0, ptr, 0, buf->size);
-          }
-        break;
-      }
     case CL_COMMAND_MAP_BUFFER:
       {
         cl_mem buffer = event->mem_objs[0];
-        if (!(buffer->flags & (CL_MEM_USE_HOST_PTR | CL_MEM_ALLOC_HOST_PTR)))
-          pocl_cuda_submit_map_mem (
-            stream,
-            cmd->map.mem_id,
-            cmd->map.mapping->offset,
-            cmd->map.mapping->size,
-            cmd->map.mapping->host_ptr);
-        POCL_LOCK_OBJ (buffer);
-        buffer->map_count++;
-        POCL_UNLOCK_OBJ (buffer);
+        pocl_cuda_submit_map_mem (
+            stream, buffer, cmd->map.mem_id, cmd->map.mapping->offset,
+            cmd->map.mapping->size, cmd->map.mapping->host_ptr);
         break;
       }
     case CL_COMMAND_UNMAP_MEM_OBJECT:
@@ -1249,13 +1316,44 @@ pocl_cuda_submit_node (_cl_command_node *node, cl_command_queue cq, int locked)
             cmd->unmap.mem_id,
             cmd->unmap.mapping->offset,
             cmd->unmap.mapping->size,
-            cmd->unmap.mapping->host_ptr);
+            cmd->unmap.mapping->host_ptr,
+            cmd->unmap.mapping->map_flags);
         break;
       }
     case CL_COMMAND_NDRANGE_KERNEL:
       pocl_cuda_submit_kernel (stream, node, node->device, node->event);
       break;
 
+    case CL_COMMAND_MIGRATE_MEM_OBJECTS:
+      switch (cmd->migrate.type)
+        {
+        case ENQUEUE_MIGRATE_TYPE_D2H:
+          {
+            cl_mem mem = event->mem_objs[0];
+            pocl_cuda_submit_read (stream, mem->mem_host_ptr,
+                                   cmd->migrate.mem_id->mem_ptr, 0, mem->size);
+            break;
+          }
+        case ENQUEUE_MIGRATE_TYPE_H2D:
+          {
+            cl_mem mem = event->mem_objs[0];
+            pocl_cuda_submit_write (stream, mem->mem_host_ptr,
+                                    cmd->migrate.mem_id->mem_ptr, 0,
+                                    mem->size);
+            break;
+          }
+        case ENQUEUE_MIGRATE_TYPE_D2D:
+          {
+            POCL_ABORT_UNIMPLEMENTED (
+                "CUDA does not support D2D migration.\n");
+          }
+        case ENQUEUE_MIGRATE_TYPE_NOP:
+          {
+            break;
+          }
+        }
+      break;
+
     case CL_COMMAND_MARKER:
     case CL_COMMAND_BARRIER:
       break;
@@ -1300,11 +1398,14 @@ void
 pocl_cuda_submit (_cl_command_node *node, cl_command_queue cq)
 {
   /* Allocate CUDA event data */
-  node->event->data
+  pocl_cuda_event_data_t *p
       = (pocl_cuda_event_data_t *)calloc (1, sizeof (pocl_cuda_event_data_t));
+  node->event->data = p;
 
   if (((pocl_cuda_queue_data_t *)cq->data)->use_threads)
     {
+
+      pthread_cond_init (&p->event_cond, NULL);
       /* Add command to work queue */
       POCL_UNLOCK_OBJ (node->event);
       pocl_cuda_queue_data_t *queue_data = (pocl_cuda_queue_data_t *)cq->data;
@@ -1368,17 +1469,7 @@ pocl_cuda_finalize_command (cl_device_id device, cl_event event)
   /* Clean up mapped memory allocations */
   if (event->command_type == CL_COMMAND_UNMAP_MEM_OBJECT)
     {
-      cl_mem buffer = event->mem_objs[0];
-      mem_mapping_t *mapping = event->command->command.unmap.mapping;
-      if (mapping->host_ptr
-          && !(buffer->flags
-               & (CL_MEM_USE_HOST_PTR | CL_MEM_ALLOC_HOST_PTR)))
-        free (mapping->host_ptr);
-
-      POCL_LOCK_OBJ (buffer);
-      DL_DELETE (buffer->mappings, mapping);
-      buffer->map_count--;
-      POCL_UNLOCK_OBJ (buffer);
+      pocl_unmap_command_finished2 (event, &event->command->command);
     }
 
   if (event->command_type == CL_COMMAND_NDRANGE_KERNEL
@@ -1399,7 +1490,7 @@ pocl_cuda_finalize_command (cl_device_id device, cl_event event)
                   if (mem->flags & CL_MEM_USE_HOST_PTR)
                     {
                       CUdeviceptr ptr
-                          = (CUdeviceptr)mem->device_ptrs[device->dev_id]
+                          = (CUdeviceptr)mem->device_ptrs[device->global_mem_id]
                                 .mem_ptr;
                       cuMemcpyDtoH (mem->mem_host_ptr, ptr, mem->size);
                       cuStreamSynchronize (0);
@@ -1475,14 +1566,29 @@ pocl_cuda_wait_event_recurse (cl_device_id device, cl_event event)
     pocl_cuda_finalize_command (device, event);
 }
 
+void
+pocl_cuda_notify_event_finished (cl_event event)
+{
+  pocl_cuda_event_data_t *e_d = (pocl_cuda_event_data_t *)event->data;
+
+  if (((pocl_cuda_queue_data_t *)event->queue->data)->use_threads)
+    pthread_cond_broadcast (&e_d->event_cond);
+}
+
 void
 pocl_cuda_wait_event (cl_device_id device, cl_event event)
 {
+  pocl_cuda_event_data_t *e_d = (pocl_cuda_event_data_t *)event->data;
+
   if (((pocl_cuda_queue_data_t *)event->queue->data)->use_threads)
     {
       /* Wait until background thread marks command as complete */
+      POCL_LOCK_OBJ (event);
       while (event->status > CL_COMPLETE)
-        ;
+        {
+          pthread_cond_wait (&e_d->event_cond, &event->pocl_lock);
+        }
+      POCL_UNLOCK_OBJ (event);
     }
   else
     {
@@ -1498,7 +1604,7 @@ pocl_cuda_free_event_data (cl_event event)
     {
       pocl_cuda_event_data_t *event_data
           = (pocl_cuda_event_data_t *)event->data;
-
+      pthread_cond_destroy (&event_data->event_cond);
       if (event->queue->properties & CL_QUEUE_PROFILING_ENABLE)
         cuEventDestroy (event_data->start);
       cuEventDestroy (event_data->end);
