@@ -491,7 +491,7 @@ void handleGetWorkDim(llvm::Module *Module, const char *KernelName) {
 
   // Clone function.
   llvm::SmallVector<llvm::ReturnInst *, 1> RI;
-  llvm::CloneFunctionInto(NewFunction, Function, VV, true, RI);
+  CloneFunctionIntoAbs(NewFunction, Function, VV, RI);
 
   Function->eraseFromParent();
 
@@ -680,7 +680,7 @@ void convertPtrArgsToOffsets(llvm::Module *Module, const char *KernelName,
       // Insert GEP to add offset.
       llvm::Value *Zero = llvm::ConstantInt::getSigned(I32ty, 0);
       llvm::GetElementPtrInst *GEP =
-          llvm::GetElementPtrInst::Create(nullptr, Base, {Zero, Offset});
+          llvm::GetElementPtrInst::Create(Base->getType()->getPointerElementType(), Base, {Zero, Offset});
 
       // Cast pointer to correct type.
       llvm::BitCastInst *Cast = new llvm::BitCastInst(GEP, ArgType);
@@ -724,7 +724,7 @@ void convertPtrArgsToOffsets(llvm::Module *Module, const char *KernelName,
 
   // Clone function.
   llvm::SmallVector<llvm::ReturnInst *, 1> RI;
-  llvm::CloneFunctionInto(NewFunction, Function, VV, true, RI);
+  CloneFunctionIntoAbs(NewFunction, Function, VV, RI);
 
   // Insert offset instructions into new function.
   for (auto Pair : ToInsert) {
