@@ -73,21 +73,19 @@
 
 #endif
 
-#ifndef HAVE_LIBDL
-#error HSA driver requires DL library
-#endif
-
 #include "pocl-hsa.h"
 #include "common.h"
+#include "common_driver.h"
 #include "devices.h"
-#include "pocl_file_util.h"
+#include "pocl-hsa.h"
 #include "pocl_cache.h"
-#include "pocl_llvm.h"
-#include "pocl_util.h"
-#include "pocl_mem_management.h"
 #include "pocl_context.h"
+#include "pocl_file_util.h"
+#include "pocl_llvm.h"
+#include "pocl_mem_management.h"
 #include "pocl_spir.h"
 #include "pocl_local_size.h"
+#include "pocl_util.h"
 
 #include <assert.h>
 #include <limits.h>
@@ -226,8 +224,7 @@ void pocl_hsa_compile_kernel_native (_cl_command_node *cmd, cl_kernel kernel,
                                      cl_device_id device, int specialize);
 
 static void*
-pocl_hsa_malloc_account(pocl_global_mem_t *mem, size_t size, hsa_region_t r,
-                        int full_profile_agent);
+pocl_hsa_malloc(pocl_global_mem_t *mem, size_t size, hsa_region_t r);
 
 void
 pocl_hsa_init_device_ops(struct pocl_device_ops *ops)
@@ -241,20 +238,28 @@ pocl_hsa_init_device_ops(struct pocl_device_ops *ops)
   ops->alloc_mem_obj = pocl_hsa_alloc_mem_obj;
   ops->free = pocl_hsa_free;
   ops->run = NULL;
-  ops->read = pocl_basic_read;
-  ops->read_rect = pocl_basic_read_rect;
-  ops->write = pocl_basic_write;
-  ops->write_rect = pocl_basic_write_rect;
-  ops->map_mem = pocl_basic_map_mem;
-  ops->unmap_mem = pocl_basic_unmap_mem;
-  ops->memfill = pocl_basic_memfill;
+
+  ops->read = pocl_driver_read;
+  ops->read_rect = pocl_driver_read_rect;
+  ops->write = pocl_driver_write;
+  ops->write_rect = pocl_driver_write_rect;
+  ops->map_mem = pocl_driver_map_mem;
+  ops->unmap_mem = pocl_driver_unmap_mem;
+  ops->get_mapping_ptr = pocl_driver_get_mapping_ptr;
+  ops->free_mapping_ptr = pocl_driver_free_mapping_ptr;
+  ops->memfill = pocl_driver_memfill;
   ops->copy = pocl_hsa_copy;
+  ops->copy_rect = pocl_driver_copy_rect;
   ops->compute_local_size = pocl_default_local_size_optimizer;
 
+  ops->get_device_info_ext = NULL;
+
   ops->svm_free = pocl_hsa_svm_free;
   ops->svm_alloc = pocl_hsa_svm_alloc;
   ops->svm_copy = pocl_hsa_svm_copy;
   ops->svm_fill = pocl_basic_svm_fill;
+  ops->svm_register = pocl_hsa_svm_register;
+  ops->svm_unregister = pocl_hsa_svm_unregister;
 
   // new driver api (out-of-order)
   ops->submit = pocl_hsa_submit;
@@ -263,11 +268,20 @@ pocl_hsa_init_device_ops(struct pocl_device_ops *ops)
   ops->notify = pocl_hsa_notify;
   ops->broadcast = pocl_hsa_broadcast;
   ops->wait_event = pocl_hsa_wait_event;
+
+  ops->build_source = pocl_driver_build_source;
+  ops->link_program = pocl_driver_link_program;
+  ops->build_binary = pocl_driver_build_binary;
+  ops->free_program = pocl_driver_free_program;
+  ops->setup_metadata = pocl_driver_setup_metadata;
+  ops->supports_binary = pocl_driver_supports_binary;
+  ops->build_poclbinary = pocl_driver_build_poclbinary;
 #if HSAIL_ENABLED
   ops->compile_kernel = pocl_hsa_compile_kernel_hsail;
 #else
   ops->compile_kernel = pocl_hsa_compile_kernel_native;
 #endif
+
   ops->update_event = pocl_hsa_update_event;
   ops->notify_event_finished = pocl_hsa_notify_event_finished;
   ops->free_event_data = pocl_hsa_free_event_data;
@@ -625,7 +639,7 @@ init_dev_data (cl_device_id dev, int count)
 
   HSA_CHECK (hsa_region_get_info (
       d->global_region, HSA_REGION_INFO_RUNTIME_ALLOC_ALIGNMENT, &sizearg));
-  dev->mem_base_addr_align = sizearg;
+  dev->mem_base_addr_align = max (sizearg, MAX_EXTENDED_ALIGNMENT);
 
   HSA_CHECK (hsa_agent_get_info (d->agent, HSA_AGENT_INFO_PROFILE,
                                  &d->agent_profile));
@@ -636,17 +650,18 @@ init_dev_data (cl_device_id dev, int count)
   dev->profile = "FULL_PROFILE";
   dev->has_own_timer = CL_TRUE;
 
+  dev->compiler_available = CL_TRUE;
+  dev->linker_available = CL_TRUE;
+
   dev->profiling_timer_resolution = (size_t) (d->timestamp_unit) || 1;
 
   if (dev->device_side_printf)
     {
-      d->printf_buffer = pocl_hsa_malloc_account
-        (dev->global_memory, dev->printf_buffer_size, d->global_region,
-         d->agent_profile == HSA_PROFILE_FULL);
+      d->printf_buffer = pocl_hsa_malloc
+        (dev->global_memory, dev->printf_buffer_size, d->global_region);
 
-      d->printf_write_pos = pocl_hsa_malloc_account
-        (dev->global_memory, sizeof (size_t), d->global_region,
-         d->agent_profile == HSA_PROFILE_FULL);
+      d->printf_write_pos = pocl_hsa_malloc
+        (dev->global_memory, sizeof (size_t), d->global_region);
     }
 
   d->exit_driver_thread = 0;
@@ -790,7 +805,7 @@ pocl_hsa_init (unsigned j, cl_device_id dev, const char *parameters)
           agent, HSA_EXT_AGENT_INFO_MAX_SAMPLER_HANDLERS, &dev->max_samplers));
     }
 
-  dev->should_allocate_svm = 1;
+  dev->svm_allocation_priority = 2;
   /* OpenCL 2.0 properties */
   dev->svm_caps = CL_DEVICE_SVM_COARSE_GRAIN_BUFFER
                   | CL_DEVICE_SVM_FINE_GRAIN_BUFFER
@@ -806,36 +821,23 @@ pocl_hsa_init (unsigned j, cl_device_id dev, const char *parameters)
   dev->on_dev_queue_props
       = CL_QUEUE_OUT_OF_ORDER_EXEC_MODE_ENABLE | CL_QUEUE_PROFILING_ENABLE;
   dev->on_host_queue_props = CL_QUEUE_PROFILING_ENABLE;
-  dev->global_mem_id = 0;
 
   pocl_hsa_device_data_t *d = init_dev_data (dev, j);
 
+  /* 0 is the host memory shared with all drivers that use it */
+  if (d->agent_profile == HSA_PROFILE_FULL)
+    dev->global_mem_id = 0;
+
   return CL_SUCCESS;
 }
 
-static void*
-pocl_hsa_malloc_account(pocl_global_mem_t *mem, size_t size, hsa_region_t r,
-                        int full_profile_agent)
-{
-/* With full profile agents, we must account for other allocations from
-   the the same virtual memory space. With base profile, we can assume the
-   global memory chunk reported by the HSA runtime is an isolated area. */
 
+static void*
+pocl_hsa_malloc (pocl_global_mem_t *mem, size_t size, hsa_region_t r)
+{
   void *b = NULL;
-  if (full_profile_agent)
-    {
-      if (mem->total_alloc_limit - mem->currently_allocated < size)
-        {
-          POCL_MSG_PRINT_INFO ("total alloc limit reached!");
-          return NULL;
-        }
-      /* FIXME: Not thread safe! */
-      mem->currently_allocated += size;
-      if (mem->max_ever_allocated < mem->currently_allocated)
-        mem->max_ever_allocated = mem->currently_allocated;
-      assert(mem->currently_allocated <= mem->total_alloc_limit);
-    }
-  else if (hsa_memory_allocate(r, size, &b) != HSA_STATUS_SUCCESS)
+
+  if (hsa_memory_allocate(r, size, &b) != HSA_STATUS_SUCCESS)
     {
       POCL_MSG_PRINT_INFO ("hsa_memory_allocate failed");
       return NULL;
@@ -854,81 +856,6 @@ pocl_hsa_malloc_account(pocl_global_mem_t *mem, size_t size, hsa_region_t r,
   return b;
 }
 
-static void *
-pocl_hsa_malloc (cl_device_id device, cl_mem_flags flags, size_t size,
-                 void *host_ptr)
-{
-  pocl_hsa_device_data_t* d = device->data;
-  void *b = NULL;
-  pocl_global_mem_t *mem = device->global_memory;
-
-  if (flags & CL_MEM_USE_HOST_PTR)
-    {
-      assert(host_ptr != NULL);
-      if (d->agent_profile == HSA_PROFILE_FULL)
-        {
-          POCL_MSG_PRINT_INFO
-            ("HSA: CL_MEM_USE_HOST_PTR FULL profile: hsa_memory_register()\n");
-          /* TODO bookkeeping of mem registrations. */
-          hsa_memory_register(host_ptr, size);
-          return host_ptr;
-        }
-      else
-        {
-          POCL_MSG_PRINT_INFO
-            ("HSA: CL_MEM_USE_HOST_PTR BASE profile: cached device copy\n");
-          return pocl_hsa_malloc_account
-              (mem, size, d->global_region,
-               d->agent_profile == HSA_PROFILE_FULL);
-        }
-    }
-
-  if (flags & CL_MEM_COPY_HOST_PTR)
-    {
-      POCL_MSG_PRINT_INFO("HSA: hsa_memory_allocate + hsa_memory_copy"
-                          " (CL_MEM_COPY_HOST_PTR)\n");
-      assert(host_ptr != NULL);
-
-      void *b = NULL;
-      /* See above (*). */
-      b = pocl_hsa_malloc_account(mem, size, d->global_region,
-                                  d->agent_profile == HSA_PROFILE_FULL);
-      if (b)
-        hsa_memory_copy(b, host_ptr, size);
-      return b;
-    }
-
-  assert(host_ptr == NULL);
-  //POCL_MSG_PRINT_INFO("HSA: hsa_memory_allocate (ALLOC_HOST_PTR)\n");
-  return pocl_hsa_malloc_account(mem, size, d->global_region,
-                                 d->agent_profile == HSA_PROFILE_FULL);
-}
-
-void
-pocl_hsa_free (cl_device_id device, cl_mem memobj)
-{
-  cl_mem_flags flags = memobj->flags;
-  void* ptr = memobj->device_ptrs[device->dev_id].mem_ptr;
-  size_t size = memobj->size;
-
-  if (flags & CL_MEM_USE_HOST_PTR ||
-      memobj->shared_mem_allocation_owner != device)
-    hsa_memory_deregister(ptr, size);
-  else
-    {
-      pocl_hsa_device_data_t *d = (pocl_hsa_device_data_t*)device->data;
-      if (d->agent_profile == HSA_PROFILE_FULL)
-        {
-          pocl_global_mem_t *mem = device->global_memory;
-          assert(mem->currently_allocated >= size);
-          mem->currently_allocated -= size;
-        }
-      hsa_memory_free(ptr);
-    }
-  if (memobj->flags | CL_MEM_ALLOC_HOST_PTR)
-    memobj->mem_host_ptr = NULL;
-}
-
 void
 pocl_hsa_copy (void *data,
                pocl_mem_identifier * dst_mem_id,
@@ -950,48 +877,52 @@ pocl_hsa_copy (void *data,
 cl_int
 pocl_hsa_alloc_mem_obj (cl_device_id device, cl_mem mem_obj, void *host_ptr)
 {
-  void *b = NULL;
-  cl_mem_flags flags = mem_obj->flags;
-  unsigned i;
+  /* if we share global memory with CPU, let the CPU driver allocate it */
+  if (device->global_mem_id == 0)
+    return pocl_basic_alloc_mem_obj (device, mem_obj, host_ptr);
 
-  /* Check if some driver has already allocated memory for this mem_obj
-     in our global address space, and use that. */
-  for (i = 0; i < mem_obj->context->num_devices; ++i)
-    {
-      if (!mem_obj->device_ptrs[i].available)
-        continue;
+  /* ... otherwise allocate it via HSA. */
+  pocl_mem_identifier *p = &mem_obj->device_ptrs[device->global_mem_id];
+  pocl_global_mem_t *gmem = device->global_memory;
+  pocl_hsa_device_data_t* d = device->data;
+  void *b = pocl_hsa_malloc (gmem, mem_obj->size, d->global_region);
+  p->mem_ptr = b;
+  p->version = 0;
 
-      if (mem_obj->device_ptrs[i].global_mem_id == device->global_mem_id
-          && mem_obj->device_ptrs[i].mem_ptr != NULL)
-        {
-          mem_obj->device_ptrs[device->dev_id].mem_ptr =
-            mem_obj->device_ptrs[i].mem_ptr;
-          hsa_memory_register (mem_obj->device_ptrs[device->dev_id].mem_ptr,
-			       mem_obj->size);
-          POCL_MSG_PRINT_INFO ("HSA: alloc_mem_obj, use already"
-                               " allocated memory\n");
-          return CL_SUCCESS;
-        }
-    }
-
-  /* Memory for this global memory is not yet allocated -> we'll allocate it. */
-  b = pocl_hsa_malloc (device, flags, mem_obj->size, host_ptr);
   if (b == NULL)
     return CL_MEM_OBJECT_ALLOCATION_FAILURE;
-
-  /* Take ownership if not USE_HOST_PTR. */
-  if (~flags & CL_MEM_USE_HOST_PTR)
-    mem_obj->shared_mem_allocation_owner = device;
-
-  mem_obj->device_ptrs[device->dev_id].mem_ptr = b;
-
-  if (flags & CL_MEM_ALLOC_HOST_PTR)
-    mem_obj->mem_host_ptr = b;
-
-  return CL_SUCCESS;
-
+  else
+    return CL_SUCCESS;
 }
 
+void
+pocl_hsa_free (cl_device_id device, cl_mem memobj)
+{
+  /* if we share global memory with CPU, let the CPU driver free it */
+  if (device->global_mem_id == 0)
+    return pocl_basic_free (device, memobj);
+
+  /* ... otherwise free it via HSA. */
+  cl_mem_flags flags = memobj->flags;
+  pocl_mem_identifier *p = &memobj->device_ptrs[device->global_mem_id];
+  hsa_memory_free(p->mem_ptr);
+  p->mem_ptr = NULL;
+  p->version = 0;
+}
+
+void pocl_hsa_svm_register (cl_device_id dev, void *host_ptr, size_t size)
+{
+  POCL_MSG_PRINT_HSA ("hsa_memory_register()\n");
+  hsa_memory_register(host_ptr, size);
+}
+
+void pocl_hsa_svm_unregister (cl_device_id dev, void *host_ptr, size_t size)
+{
+  POCL_MSG_PRINT_HSA ("hsa_memory_deregister()\n");
+  hsa_memory_deregister(host_ptr, size);
+}
+
+
 static void
 setup_kernel_args (pocl_hsa_device_data_t *d,
                    _cl_command_node *cmd,
@@ -1056,10 +987,12 @@ setup_kernel_args (pocl_hsa_device_data_t *d,
             }
           else
             {
-              cl_mem m = *(cl_mem *)al->value;
               uint64_t dev_ptr = 0;
-              if (m->device_ptrs)
+              if (al->is_svm)
+                dev_ptr = (uint64_t) (*(void **)al->value);
+              else
                 {
+                  cl_mem m = *(cl_mem *)al->value;
                   dev_ptr
                       = (uint64_t)m->device_ptrs[cmd->device->dev_id].mem_ptr;
                   if (m->flags & CL_MEM_USE_HOST_PTR
@@ -1074,8 +1007,6 @@ setup_kernel_args (pocl_hsa_device_data_t *d,
                                        m->size);
                     }
                 }
-              else
-                dev_ptr = (uint64_t)m->mem_host_ptr;
 
               dev_ptr += al->offset;
               memcpy (write_pos, &dev_ptr, sizeof (uint64_t));
@@ -1114,9 +1045,9 @@ setup_kernel_args (pocl_hsa_device_data_t *d,
   /* Copy the context object to HSA allocated global memory to ensure Base
      profile agents can access it. */
 
-  event_data->context = pocl_hsa_malloc_account
+  event_data->context = pocl_hsa_malloc
     (d->device->global_memory, POCL_CONTEXT_SIZE (d->device->address_bits),
-     d->global_region, d->agent_profile == HSA_PROFILE_FULL);
+     d->global_region);
 
   if (d->device->address_bits == 64)
     memcpy (event_data->context, &cmd->command.run.pc, sizeof (struct pocl_context));
@@ -1589,7 +1520,7 @@ pocl_hsa_submit (_cl_command_node *node, cl_command_queue cq)
   else
     PN_ADD(d->wait_list, node->event);
 
-  POCL_MSG_PRINT_INFO("After Event %u submit: WL : %li, RL: %li\n",
+  POCL_MSG_PRINT_INFO("After Event %" PRIu64 " submit: WL : %li, RL: %li\n",
                       node->event->id, d->wait_list_size, d->ready_list_size);
 
   POCL_UNLOCK_OBJ (node->event);
@@ -1617,11 +1548,11 @@ pocl_hsa_join (cl_device_id device, cl_command_queue cq)
   POCL_RETAIN_OBJECT_UNLOCKED (event);
   POCL_UNLOCK_OBJ (cq);
 
-  POCL_MSG_PRINT_HSA ("device->join on event %u\n", event->id);
+  POCL_MSG_PRINT_HSA ("device->join on event %" PRIu64 "\n", event->id);
 
   if (event->status <= CL_COMPLETE)
     {
-      POCL_MSG_PRINT_HSA ("device->join: last event (%u) in queue"
+      POCL_MSG_PRINT_HSA ("device->join: last event (%" PRIu64 ") in queue"
                           " exists, but is complete\n", event->id);
       goto RETURN;
     }
@@ -1631,7 +1562,7 @@ pocl_hsa_join (cl_device_id device, cl_command_queue cq)
       pocl_hsa_event_data_t *e_d = (pocl_hsa_event_data_t *)event->data;
       PTHREAD_CHECK (pthread_cond_wait (&e_d->event_cond, &event->pocl_lock));
     }
-  POCL_MSG_PRINT_HSA ("device->join on event %u finished"
+  POCL_MSG_PRINT_HSA ("device->join on event %" PRIu64 " finished"
                       " with status: %i\n", event->id, event->status);
 
 RETURN:
@@ -1654,7 +1585,7 @@ pocl_hsa_notify (cl_device_id device, cl_event event, cl_event finished)
   pocl_hsa_device_data_t *d = device->data;
   _cl_command_node *node = event->command;
   int added_to_readylist = 0;
-  POCL_MSG_PRINT_HSA ("notify on event %u \n", event->id);
+  POCL_MSG_PRINT_HSA ("notify on event %" PRIu64 " \n", event->id);
 
   if (finished->status < CL_COMPLETE)
     {
@@ -1678,19 +1609,19 @@ pocl_hsa_notify (cl_device_id device, cl_event event, cl_event finished)
               break;
           if (i < d->wait_list_size)
             {
-              POCL_MSG_PRINT_INFO("event %u wait_list -> ready_list\n", 
+              POCL_MSG_PRINT_INFO("event %" PRIu64 " wait_list -> ready_list\n",
                                   event->id);
               PN_ADD(d->ready_list, event);
               PN_REMOVE(d->wait_list, i);
             }
           else
-            POCL_ABORT("cant move event %u from waitlist to"
+            POCL_ABORT("cant move event %" PRIu64 " from waitlist to"
                        " readylist - not found in waitlist\n", event->id);
           added_to_readylist = 1;
           PTHREAD_CHECK(pthread_mutex_unlock(&d->list_mutex));
         }
       else
-        POCL_MSG_WARN ("node->ready was 1 but event %u is"
+        POCL_MSG_WARN ("node->ready was 1 but event %" PRIu64 " is"
                        " not queued: status %i!\n",
                        event->id, event->status);
     }
@@ -1709,12 +1640,12 @@ pocl_hsa_broadcast (cl_event event)
 void
 pocl_hsa_wait_event(cl_device_id device, cl_event event)
 {
-  POCL_MSG_PRINT_HSA ("device->wait_event on event %u\n", event->id);
+  POCL_MSG_PRINT_HSA ("device->wait_event on event %" PRIu64 "\n", event->id);
   POCL_LOCK_OBJ (event);
   if (event->status <= CL_COMPLETE)
     {
       POCL_MSG_PRINT_HSA ("device->wain_event: last event"
-                          " (%u) in queue exists, but is complete\n", 
+                          " (%" PRIu64 ") in queue exists, but is complete\n",
                           event->id);
       POCL_UNLOCK_OBJ(event);
       return;
@@ -1897,7 +1828,7 @@ pocl_hsa_ndrange_event_finished (pocl_hsa_device_data_t *d, size_t i)
   POCL_LOCK_OBJ (event);
   pocl_hsa_event_data_t *event_data = (pocl_hsa_event_data_t *)event->data;
 
-  POCL_MSG_PRINT_INFO("event %u finished, removing from running_list\n",
+  POCL_MSG_PRINT_INFO("event %" PRIu64 " finished, removing from running_list\n",
                       event->id);
   dd->running_events[i] = dd->running_events[--dd->running_list_size];
 
@@ -1963,12 +1894,12 @@ pocl_hsa_run_ready_commands (pocl_hsa_device_data_t *d)
               e->command, e->command->command.run.kernel, e->queue->device, 1);
           pocl_hsa_launch (d, e);
           enqueued_ndrange = 1;
-          POCL_MSG_PRINT_INFO ("NDrange event %u launched, remove"
+          POCL_MSG_PRINT_INFO ("NDrange event %" PRIu64 " launched, remove"
                                " from readylist\n", e->id);
         }
       else
         {
-          POCL_MSG_PRINT_INFO ("running non-NDrange event %u,"
+          POCL_MSG_PRINT_INFO ("running non-NDrange event %" PRIu64 ","
                                " remove from readylist\n", e->id);
           pocl_exec_command (e->command);
         }
@@ -2168,14 +2099,19 @@ pocl_hsa_svm_free (cl_device_id dev, void *svm_ptr)
 void *
 pocl_hsa_svm_alloc (cl_device_id dev, cl_svm_mem_flags flags, size_t size)
 {
-  POCL_RETURN_ERROR_ON (((flags & CL_MEM_SVM_ATOMICS)
-                         && ((dev->svm_caps & CL_DEVICE_SVM_ATOMICS) == 0)),
-                        NULL, "This device doesn't have SVM Atomics");
+  if ((flags & CL_MEM_SVM_ATOMICS)
+      && ((dev->svm_caps & CL_DEVICE_SVM_ATOMICS) == 0))
+    {
+      POCL_MSG_ERR ("This device doesn't support SVM Atomics\n");
+      return NULL;
+    }
 
-  POCL_RETURN_ERROR_ON (
-      ((flags & CL_MEM_SVM_FINE_GRAIN_BUFFER)
-       && ((dev->svm_caps & CL_DEVICE_SVM_FINE_GRAIN_BUFFER) == 0)),
-      NULL, "This device doesn't have SVM Atomics");
+  if ((flags & CL_MEM_SVM_FINE_GRAIN_BUFFER)
+       && ((dev->svm_caps & CL_DEVICE_SVM_FINE_GRAIN_BUFFER) == 0))
+    {
+      POCL_MSG_ERR ("This device doesn't support SVM Fine grained Buffer\n");
+      return NULL;
+    }
 
   pocl_hsa_device_data_t *d = (pocl_hsa_device_data_t *)dev->data;
   void *b = NULL;
