@@ -24,6 +24,8 @@
 
 #include "ttasim.h"
 #include "bufalloc.h"
+#include "common.h"
+#include "common_driver.h"
 #include "pocl_device.h"
 #include "pocl_util.h"
 #include "common.h"
@@ -90,15 +92,27 @@ pocl_ttasim_init_device_ops(struct pocl_device_ops *ops)
   ops->copy_rect = pocl_tce_copy_rect;
   ops->map_mem = pocl_tce_map_mem;
   ops->unmap_mem = pocl_tce_unmap_mem;
+  ops->get_mapping_ptr = pocl_driver_get_mapping_ptr;
+  ops->free_mapping_ptr = pocl_driver_free_mapping_ptr;
   ops->run = pocl_tce_run;
   ops->init_build = pocl_tce_init_build;
   ops->flush = pocl_tce_flush;
   ops->join = pocl_tce_join;
   ops->submit = pocl_tce_submit;
+
+  ops->build_source = pocl_driver_build_source;
+  ops->link_program = pocl_driver_link_program;
+  ops->build_binary = pocl_driver_build_binary;
+  ops->free_program = pocl_driver_free_program;
+  ops->setup_metadata = pocl_driver_setup_metadata;
+  ops->supports_binary = pocl_driver_supports_binary;
+  ops->build_poclbinary = pocl_driver_build_poclbinary;
   ops->compile_kernel = pocl_tce_compile_kernel;
+
   ops->broadcast = pocl_broadcast;
   ops->notify = pocl_tce_notify;
   ops->build_hash = pocl_tce_build_hash;
+  ops->get_device_info_ext = NULL;
 }
 
 
@@ -549,7 +563,10 @@ pocl_ttasim_init (unsigned j, cl_device_id dev, const char* parameters)
   dev->host_unified_memory = CL_FALSE;
 
   dev->available = CL_TRUE;
+#ifdef ENABLE_LLVM
   dev->compiler_available = CL_TRUE;
+  dev->linker_available = CL_TRUE;
+#endif
   dev->spmd = CL_FALSE;
   dev->workgroup_pass = CL_TRUE;
   dev->execution_capabilities = CL_EXEC_KERNEL;
@@ -578,6 +595,9 @@ pocl_ttasim_init (unsigned j, cl_device_id dev, const char* parameters)
   dev->num_partition_types = 0;
   dev->partition_type = NULL;
 
+  dev->mem_base_addr_align =
+      dev->min_data_type_align_size = 16;
+
   new TTASimDevice(dev, parameters);
   return CL_SUCCESS;
 }
@@ -607,7 +627,7 @@ void pocl_ttasim_update_event (cl_device_id device, cl_event event, cl_int statu
         event->time_start = d->timeStamp();
       break;
     case CL_COMPLETE:
-      POCL_MSG_PRINT_INFO("TTA: Command complete, event %d\n", event->id);
+      POCL_MSG_PRINT_INFO("TTA: Command complete, event %zu\n", event->id);
       POCL_LOCK_OBJ (event);
       if (event->queue->properties & CL_QUEUE_PROFILING_ENABLE)
         event->time_end = d->timeStamp();
