@@ -17,6 +17,8 @@
 
 #include "config.h"
 
+#include "pocl_export.h"
+
 // size_t print spec
 #ifndef PRIuS
 # define PRIuS "zu"
@@ -57,8 +59,10 @@ extern "C" {
 #define POCL_DEBUG_FLAG_HSA 0x100
 #define POCL_DEBUG_FLAG_TCE 0x200
 #define POCL_DEBUG_FLAG_CUDA 0x400
-#define POCL_DEBUG_FLAG_WARNING 0x800
-#define POCL_DEBUG_FLAG_ERROR 0x1000
+#define POCL_DEBUG_FLAG_ACCEL 0x800
+
+#define POCL_DEBUG_FLAG_WARNING  0x8000000000UL
+#define POCL_DEBUG_FLAG_ERROR   0x10000000000UL
 #define POCL_DEBUG_FLAG_ALL (uint64_t)(-1)
 
 #define POCL_FILTER_TYPE_INFO 1
@@ -103,28 +107,25 @@ extern "C" {
                               *errcode_ret = CL_SUCCESS;                \
                             } } while (0)
 
-
-
-#include "config.h"
-
 #ifdef POCL_DEBUG_MESSAGES
 
+POCL_EXPORT
     extern uint64_t pocl_debug_messages_filter;
-    extern int stderr_is_a_tty;
+POCL_EXPORT
+    extern int pocl_stderr_is_a_tty;
 
     #define POCL_DEBUGGING_ON (pocl_debug_messages_filter)
 
-    #if __GNUC__ >= 2
-    #define __func__ __PRETTY_FUNCTION__
-    #else
-    #define __func__ __FUNCTION__
-    #endif
-
         #define POCL_DEBUG_HEADER(FILTER, FILTER_TYPE) \
             pocl_debug_print_header (__func__, __LINE__, #FILTER, FILTER_TYPE);
-        extern void pocl_debug_output_lock ();
-        extern void pocl_debug_output_unlock ();
+
+POCL_EXPORT
+        extern void pocl_debug_output_lock (void);
+POCL_EXPORT
+        extern void pocl_debug_output_unlock (void);
+POCL_EXPORT
         extern void pocl_debug_messages_setup (const char *debug);
+POCL_EXPORT
         extern void pocl_debug_print_header (const char * func, unsigned line,
                                              const char* filter, int filter_type);
         extern void pocl_debug_measure_start (uint64_t* start);
@@ -148,7 +149,7 @@ extern "C" {
           if (pocl_debug_messages_filter & POCL_DEBUG_FLAG_ ## FILTER) {    \
             pocl_debug_output_lock ();                                      \
                 POCL_DEBUG_HEADER(FILTER, POCL_FILTER_TYPE_ ## TYPE)        \
-                if (stderr_is_a_tty)                                        \
+                if (pocl_stderr_is_a_tty)                                   \
                   fprintf (stderr, "%s", POCL_COLOR_BOLDRED                 \
                                     ERRCODE " "  POCL_COLOR_RESET);         \
                 else                                                        \
@@ -185,6 +186,8 @@ extern "C" {
     #define POCL_MSG_PRINT_INFO_F(filter, errcode, ...) \
           POCL_MSG_PRINT_F(filter, INFO, errcode, __VA_ARGS__)
 
+    #define POCL_MSG_PRINT_ACCEL2(errcode, ...) POCL_MSG_PRINT_INFO_F(ACCEL, errcode, __VA_ARGS__)
+    #define POCL_MSG_PRINT_ACCEL(...) POCL_MSG_PRINT_INFO_F(ACCEL, "", __VA_ARGS__)
     #define POCL_MSG_PRINT_CUDA2(errcode, ...) POCL_MSG_PRINT_INFO_F(CUDA, errcode, __VA_ARGS__)
     #define POCL_MSG_PRINT_CUDA(...) POCL_MSG_PRINT_INFO_F(CUDA, "", __VA_ARGS__)
     #define POCL_MSG_PRINT_HSA2(errcode, ...) POCL_MSG_PRINT_INFO_F(HSA, errcode, __VA_ARGS__)
@@ -226,6 +229,8 @@ extern "C" {
     #define POCL_MEASURE_FINISH(...)  do {} while (0)
     #define POCL_DEBUG_EVENT_TIME(...)  do {} while (0)
 
+    #define POCL_MSG_PRINT_ACCEL2(...)  do {} while (0)
+    #define POCL_MSG_PRINT_ACCEL(...)  do {} while (0)
     #define POCL_MSG_PRINT_CUDA2(...)  do {} while (0)
     #define POCL_MSG_PRINT_CUDA(...)  do {} while (0)
     #define POCL_MSG_PRINT_HSA2(...)  do {} while (0)
