@@ -90,9 +90,6 @@ using namespace llvm;
 static std::map<cl_device_id, llvm::TargetMachine *> targetMachines;
 static std::map<cl_device_id, PassManager *> kernelPasses;
 
-// This is used to control the kernel we process in the kernel compilation.
-extern cl::opt<std::string> KernelName;
-
 /* FIXME: these options should come from the cl_device, and
  * cl_program's options. */
 static llvm::TargetOptions GetTargetOptions() {
@@ -163,9 +160,7 @@ static TargetMachine *GetTargetMachine(cl_device_id device, Triple &triple) {
 }
 /* helpers copied from LLVM opt END */
 
-static PassManager &
-kernel_compiler_passes(cl_device_id device, llvm::Module *input,
-                       const std::string &module_data_layout) {
+static PassManager &kernel_compiler_passes(cl_device_id device) {
 
   PassManager *Passes = nullptr;
   PassRegistry *Registry = nullptr;
@@ -220,6 +215,7 @@ kernel_compiler_passes(cl_device_id device, llvm::Module *input,
      PHI). */
 
   std::vector<std::string> passes;
+  passes.push_back("inline-kernels");
   passes.push_back("remove-optnone");
   passes.push_back("optimize-wi-func-calls");
   passes.push_back("handle-samplers");
@@ -345,101 +341,144 @@ kernel_compiler_passes(cl_device_id device, llvm::Module *input,
   return *Passes;
 }
 
-void pocl_destroy_llvm_module(void *modp) {
+void pocl_destroy_llvm_module(void *modp, cl_context ctx) {
 
-  PoclCompilerMutexGuard lockHolder(NULL);
-  InitializeLLVM();
+  PoclLLVMContextData *llvm_ctx = (PoclLLVMContextData *)ctx->llvm_context_data;
+  PoclCompilerMutexGuard lockHolder(&llvm_ctx->Lock);
 
   llvm::Module *mod = (llvm::Module *)modp;
   if (mod) {
     delete mod;
-    --numberOfIRs;
+    --llvm_ctx->number_of_IRs;
   }
 }
 
 // The global variables used to control the WG function generation's
 // specialization parameteres. Defined in lib/llvmopencl/WorkitemHandler.cc.
 namespace pocl {
-extern size_t WGLocalSizeX;
-extern size_t WGLocalSizeY;
-extern size_t WGLocalSizeZ;
-extern bool WGDynamicLocalSize;
-extern size_t WGMaxGridDimWidth;
-extern bool WGAssumeZeroGlobalOffset;
 }
 
-int pocl_update_program_llvm_irs_unlocked(cl_program program,
-                                          unsigned device_i);
-
 int pocl_llvm_generate_workgroup_function_nowrite(
     unsigned DeviceI, cl_device_id Device, cl_kernel Kernel,
     _cl_command_node *Command, void **Output, int Specialize) {
 
   _cl_command_run *RunCommand = &Command->command.run;
   cl_program Program = Kernel->program;
+  cl_context ctx = Program->context;
+  PoclLLVMContextData *llvm_ctx = (PoclLLVMContextData *)ctx->llvm_context_data;
+  PoclCompilerMutexGuard lockHolder(&llvm_ctx->Lock);
 
-  currentPoclDevice = Device;
+#ifdef DEBUG_POCL_LLVM_API
+  printf("### calling the kernel compiler for kernel %s local_x %zu "
+         "local_y %zu local_z %zu parallel_filename: %s\n",
+         kernel->name, local_x, local_y, local_z, parallel_bc_path);
+#endif
+  assert(Program->data[DeviceI] != nullptr);
 
-  PoclCompilerMutexGuard LockHolder(NULL);
-  InitializeLLVM();
-
-  if (Program->llvm_irs[DeviceI] == NULL)
-    pocl_update_program_llvm_irs_unlocked(Program, DeviceI);
-
-  llvm::Module *ProgramBC = (llvm::Module *)Program->llvm_irs[DeviceI];
+  llvm::Module *ProgramBC = (llvm::Module *)Program->data[DeviceI];
 
   // Create an empty Module and copy only the kernel+callgraph from
   // program.bc.
   llvm::Module *ParallelBC =
-      new llvm::Module(StringRef("parallel_bc"), GlobalContext());
+      new llvm::Module(StringRef("parallel_bc"), *llvm_ctx->Context);
 
   ParallelBC->setTargetTriple(ProgramBC->getTargetTriple());
   ParallelBC->setDataLayout(ProgramBC->getDataLayout());
 
-  copyKernelFromBitcode(Kernel->name, ParallelBC, ProgramBC);
+  copyKernelFromBitcode(Kernel->name, ParallelBC, ProgramBC,
+                        Device->global_as_id, Device->device_aux_functions);
+
+  // Set to true to generate a global offset 0 specialized WG function.
+  bool WGAssumeZeroGlobalOffset;
+  // If set to true, the next 3 parameters define the local size to specialize
+  // for.
+  bool WGDynamicLocalSize;
+  size_t WGLocalSizeX;
+  size_t WGLocalSizeY;
+  size_t WGLocalSizeZ;
+  // If set to non-zero, assume each grid dimension is at most this
+  // work-items wide.
+  size_t WGMaxGridDimWidth;
 
   // Set the specialization properties.
   if (Specialize) {
-    pocl::WGLocalSizeX = RunCommand->pc.local_size[0];
-    pocl::WGLocalSizeY = RunCommand->pc.local_size[1];
-    pocl::WGLocalSizeZ = RunCommand->pc.local_size[2];
-    pocl::WGDynamicLocalSize = pocl::WGLocalSizeX == 0 &&
-                               pocl::WGLocalSizeY == 0 &&
-                               pocl::WGLocalSizeZ == 0;
-    pocl::WGAssumeZeroGlobalOffset = RunCommand->pc.global_offset[0] == 0 &&
-                                     RunCommand->pc.global_offset[1] == 0 &&
-                                     RunCommand->pc.global_offset[2] == 0;
+    WGLocalSizeX = RunCommand->pc.local_size[0];
+    WGLocalSizeY = RunCommand->pc.local_size[1];
+    WGLocalSizeZ = RunCommand->pc.local_size[2];
+    WGDynamicLocalSize =
+        WGLocalSizeX == 0 && WGLocalSizeY == 0 && WGLocalSizeZ == 0;
+    WGAssumeZeroGlobalOffset = RunCommand->pc.global_offset[0] == 0 &&
+                               RunCommand->pc.global_offset[1] == 0 &&
+                               RunCommand->pc.global_offset[2] == 0;
     // Compile a smallgrid version or a generic one?
     if (RunCommand->force_large_grid_wg_func ||
         pocl_cmd_max_grid_dim_width(RunCommand) >=
             Device->grid_width_specialization_limit) {
-      pocl::WGMaxGridDimWidth = 0; // The generic / large / unlimited size one.
+      WGMaxGridDimWidth = 0; // The generic / large / unlimited size one.
     } else {
       // Limited grid dimension width by the device specific limit.
-      pocl::WGMaxGridDimWidth = Device->grid_width_specialization_limit;
+      WGMaxGridDimWidth = Device->grid_width_specialization_limit;
     }
   } else {
-    pocl::WGDynamicLocalSize = true;
-    pocl::WGLocalSizeX = pocl::WGLocalSizeY = pocl::WGLocalSizeZ = 0;
-    pocl::WGAssumeZeroGlobalOffset = false;
-    pocl::WGMaxGridDimWidth = 0;
+    WGDynamicLocalSize = true;
+    WGLocalSizeX = WGLocalSizeY = WGLocalSizeZ = 0;
+    WGAssumeZeroGlobalOffset = false;
+    WGMaxGridDimWidth = 0;
   }
 
-  KernelName = Kernel->name;
+  if (Device->device_aux_functions) {
+    std::string concat;
+    const char **tmp = Device->device_aux_functions;
+    while (*tmp != nullptr) {
+      concat.append(*tmp);
+      ++tmp;
+      if (*tmp)
+        concat.append(";");
+    }
+    setModuleStringMetadata(ParallelBC, "device_aux_functions", concat.c_str());
+  }
+
+  setModuleIntMetadata(ParallelBC, "device_address_bits", Device->address_bits);
+  setModuleBoolMetadata(ParallelBC, "device_arg_buffer_launcher",
+                        Device->arg_buffer_launcher);
+  setModuleBoolMetadata(ParallelBC, "device_is_spmd", Device->spmd);
+
+  setModuleStringMetadata(ParallelBC, "KernelName", Kernel->name);
+  setModuleIntMetadata(ParallelBC, "WGMaxGridDimWidth", WGMaxGridDimWidth);
+  setModuleIntMetadata(ParallelBC, "WGLocalSizeX", WGLocalSizeX);
+  setModuleIntMetadata(ParallelBC, "WGLocalSizeY", WGLocalSizeY);
+  setModuleIntMetadata(ParallelBC, "WGLocalSizeZ", WGLocalSizeZ);
+  setModuleBoolMetadata(ParallelBC, "WGDynamicLocalSize", WGDynamicLocalSize);
+  setModuleBoolMetadata(ParallelBC, "WGAssumeZeroGlobalOffset",
+                        WGAssumeZeroGlobalOffset);
+
+  setModuleIntMetadata(ParallelBC, "device_global_as_id", Device->global_as_id);
+  setModuleIntMetadata(ParallelBC, "device_local_as_id", Device->local_as_id);
+  setModuleIntMetadata(ParallelBC, "device_constant_as_id",
+                       Device->constant_as_id);
+  setModuleIntMetadata(ParallelBC, "device_args_as_id", Device->args_as_id);
+  setModuleIntMetadata(ParallelBC, "device_context_as_id",
+                       Device->context_as_id);
+
+  setModuleBoolMetadata(ParallelBC, "device_side_printf",
+                        Device->device_side_printf);
+  setModuleBoolMetadata(ParallelBC, "device_alloca_locals",
+                        Device->device_alloca_locals);
+
+  setModuleIntMetadata(ParallelBC, "device_max_witem_dim",
+                       Device->max_work_item_dimensions);
+  setModuleIntMetadata(ParallelBC, "device_max_witem_sizes_0",
+                       Device->max_work_item_sizes[0]);
+  setModuleIntMetadata(ParallelBC, "device_max_witem_sizes_1",
+                       Device->max_work_item_sizes[1]);
+  setModuleIntMetadata(ParallelBC, "device_max_witem_sizes_2",
+                       Device->max_work_item_sizes[2]);
 
 #ifdef DUMP_LLVM_PASS_TIMINGS
   llvm::TimePassesIsEnabled = true;
 #endif
   POCL_MEASURE_START(llvm_workgroup_ir_func_gen);
-#ifdef LLVM_OLDER_THAN_3_7
-  kernel_compiler_passes(Device, ParallelBC,
-                         ParallelBC->getDataLayout()->getStringRepresentation())
-      .run(*ParallelBC);
-#else
-  kernel_compiler_passes(Device, ParallelBC,
-                         ParallelBC->getDataLayout().getStringRepresentation())
-      .run(*ParallelBC);
-#endif
+  kernel_compiler_passes(Device).run(*ParallelBC);
   POCL_MEASURE_FINISH(llvm_workgroup_ir_func_gen);
 #ifdef DUMP_LLVM_PASS_TIMINGS
   llvm::reportAndResetTimings();
@@ -447,12 +486,12 @@ int pocl_llvm_generate_workgroup_function_nowrite(
 
   // Print loop vectorizer remarks if enabled.
   if (pocl_get_bool_option("POCL_VECTORIZER_REMARKS", 0) == 1) {
-    std::cout << getDiagString();
+    std::cout << getDiagString(ctx);
   }
 
   assert(Output != NULL);
   *Output = (void *)ParallelBC;
-  ++numberOfIRs;
+  ++llvm_ctx->number_of_IRs;
   return 0;
 }
 
@@ -460,7 +499,7 @@ int pocl_llvm_generate_workgroup_function(unsigned DeviceI, cl_device_id Device,
                                           cl_kernel Kernel,
                                           _cl_command_node *Command,
                                           int Specialize) {
-
+  cl_context ctx = Kernel->context;
   void *Module = NULL;
 
   char ParallelBCPath[POCL_FILENAME_LENGTH];
@@ -491,94 +530,93 @@ int pocl_llvm_generate_workgroup_function(unsigned DeviceI, cl_device_id Device,
     return Error;
   }
 
-  pocl_destroy_llvm_module(Module);
+  pocl_destroy_llvm_module(Module, ctx);
   return Error;
 }
 
-int pocl_update_program_llvm_irs_unlocked(cl_program program,
-                                          unsigned device_i) {
+/* Reads LLVM IR module from program->binaries[i], if program->data[device_i] is
+ * NULL */
+int pocl_llvm_read_program_llvm_irs(cl_program program, unsigned device_i,
+                                    const char *program_bc_path) {
+  cl_context ctx = program->context;
+  PoclLLVMContextData *llvm_ctx = (PoclLLVMContextData *)ctx->llvm_context_data;
+  PoclCompilerMutexGuard lockHolder(&llvm_ctx->Lock);
 
-  char program_bc_path[POCL_FILENAME_LENGTH];
-  pocl_cache_program_bc_path(program_bc_path, program, device_i);
+  if (program->data[device_i] != nullptr)
+    return CL_SUCCESS;
 
-  if (!pocl_exists(program_bc_path))
-    {
-      POCL_MSG_ERR ("%s does not exist!\n",
-                     program_bc_path);
-      return -1;
-    }
-
-  assert(program->llvm_irs[device_i] == nullptr);
-  program->llvm_irs[device_i] = parseModuleIR(program_bc_path);
-  ++numberOfIRs;
-  return 0;
+  if (program->binaries[device_i])
+    program->data[device_i] =
+        parseModuleIRMem((char *)program->binaries[device_i],
+                         program->binary_sizes[device_i], llvm_ctx->Context);
+  else {
+    // TODO
+    assert(program_bc_path);
+    program->data[device_i] = parseModuleIR(program_bc_path, llvm_ctx->Context);
+  }
+  assert(program->data[device_i]);
+  ++llvm_ctx->number_of_IRs;
+  return CL_SUCCESS;
 }
 
-int pocl_update_program_llvm_irs(cl_program program,
-                                 unsigned device_i) {
-  PoclCompilerMutexGuard lockHolder(NULL);
-  InitializeLLVM();
-
-  return pocl_update_program_llvm_irs_unlocked(program, device_i);
-}
-
-void pocl_free_llvm_irs(cl_program program, unsigned device_i) {
-  if (program->llvm_irs[device_i]) {
-    PoclCompilerMutexGuard lockHolder(NULL);
-    InitializeLLVM();
-    llvm::Module *mod = (llvm::Module *)program->llvm_irs[device_i];
+void pocl_llvm_free_llvm_irs(cl_program program, unsigned device_i) {
+  cl_context ctx = program->context;
+  PoclLLVMContextData *llvm_ctx = (PoclLLVMContextData *)ctx->llvm_context_data;
+  PoclCompilerMutexGuard lockHolder(&llvm_ctx->Lock);
+  if (program->data[device_i]) {
+    llvm::Module *mod = (llvm::Module *)program->data[device_i];
     delete mod;
-    --numberOfIRs;
-    program->llvm_irs[device_i] = NULL;
+    --llvm_ctx->number_of_IRs;
+    program->data[device_i] = nullptr;
   }
 }
 
-void pocl_llvm_update_binaries(cl_program program) {
+/* writes fresh program.bc and program->binaries[i] from LLVM IR module. */
+int pocl_llvm_update_binaries(cl_program program, cl_uint device_i) {
 
-  PoclCompilerMutexGuard lockHolder(NULL);
-  InitializeLLVM();
+  cl_context ctx = program->context;
+  PoclLLVMContextData *llvm_ctx = (PoclLLVMContextData *)ctx->llvm_context_data;
+  PoclCompilerMutexGuard lockHolder(&llvm_ctx->Lock);
 
   char program_bc_path[POCL_FILENAME_LENGTH];
   int error;
 
   // Dump the LLVM IR Modules to memory buffers.
-  assert(program->llvm_irs != NULL);
+  assert(program->data != NULL);
 #ifdef DEBUG_POCL_LLVM_API
   printf("### refreshing the binaries of the program %p\n", program);
 #endif
 
-  for (size_t i = 0; i < program->num_devices; ++i) {
-    assert(program->llvm_irs[i] != NULL);
-    if (program->binaries[i])
-      continue;
+  cl_uint i = device_i;
+  // LLVM IR
+  assert(program->data[i] != NULL);
 
-    pocl_cache_program_bc_path(program_bc_path, program, i);
-    error = pocl_write_module((llvm::Module *)program->llvm_irs[i],
-                              program_bc_path, 1);
-    assert(error == 0);
-    if (error)
-      {
-        POCL_MSG_ERR ("pocl_write_module(%s) failed!\n",
-                     program_bc_path);
-        continue;
-      }
+  POCL_MEM_FREE(program->binaries[i]);
 
-    std::string content;
-    writeModuleIR((llvm::Module *)program->llvm_irs[i], content);
+  pocl_cache_program_bc_path(program_bc_path, program, i);
+  error =
+      pocl_write_module((llvm::Module *)program->data[i], program_bc_path, 1);
+  assert(error == 0);
+  if (error) {
+    POCL_MSG_ERR("pocl_write_module(%s) failed!\n", program_bc_path);
+    return error;
+  }
 
-    size_t n = content.size();
-    if (n < program->binary_sizes[i])
-      POCL_ABORT("binary size doesn't match the expected value\n");
-    if (program->binaries[i])
-      POCL_MEM_FREE(program->binaries[i]);
-    program->binaries[i] = (unsigned char *)malloc(n);
-    std::memcpy(program->binaries[i], content.c_str(), n);
+  std::string content;
+  writeModuleIR((llvm::Module *)program->data[i], content);
+
+  size_t n = content.size();
+  if (n < program->binary_sizes[i])
+    POCL_ABORT("binary size doesn't match the expected value\n");
+
+  program->binaries[i] = (unsigned char *)malloc(n);
+  std::memcpy(program->binaries[i], content.c_str(), n);
 
 #ifdef DEBUG_POCL_LLVM_API
-    printf("### binary for device %zi was of size %zu\n", i,
-           program->binary_sizes[i]);
+  printf("### binary for device %zi was of size %zu\n", i,
+         program->binary_sizes[i]);
 #endif
-  }
+  return CL_SUCCESS;
 }
 
 static void initPassManagerForCodeGen(PassManager& PM, cl_device_id Device) {
@@ -593,11 +631,12 @@ static void initPassManagerForCodeGen(PassManager& PM, cl_device_id Device) {
 /* Run LLVM codegen on input file (parallel-optimized).
  * modp = llvm::Module* of parallel.bc
  * Output native object file (<kernel>.so.o). */
-int pocl_llvm_codegen(cl_device_id Device, void *Modp, char **Output,
-                      uint64_t *OutputSize) {
+int pocl_llvm_codegen(cl_device_id Device, cl_program program, void *Modp,
+                      char **Output, uint64_t *OutputSize) {
 
-  PoclCompilerMutexGuard LockHolder(nullptr);
-  InitializeLLVM();
+  cl_context ctx = program->context;
+  PoclLLVMContextData *llvm_ctx = (PoclLLVMContextData *)ctx->llvm_context_data;
+  PoclCompilerMutexGuard lockHolder(&llvm_ctx->Lock);
 
   llvm::Module *Input = (llvm::Module *)Modp;
   assert(Input);
