@@ -28,15 +28,15 @@
 
 #include <time.h>
 
-#ifndef _MSC_VER
+#ifndef _WIN32
 #include <dirent.h>
+#include <string.h>
 #include <sys/resource.h>
 #include <sys/stat.h>
 #include <sys/time.h>
 #include <sys/types.h>
 #include <sys/wait.h>
 #include <unistd.h>
-#include <unistd.h>
 #include <utime.h>
 #else
 #  include "vccompat.hpp"
@@ -207,7 +207,7 @@ pocl_size_ceil2(size_t x) {
   /* Rounds up to the next highest power of two without branching and
    * is as fast as a BSR instruction on x86, see:
    *
-   * http://graphics.stanford.edu/~seander/bithacks.html#RoundUpPowerOf2
+   * https://graphics.stanford.edu/~seander/bithacks.html#RoundUpPowerOf2
    */
   --x;
   x |= x >> 1;
@@ -227,7 +227,7 @@ pocl_size_ceil2_64 (uint64_t x)
   /* Rounds up to the next highest power of two without branching and
    * is as fast as a BSR instruction on x86, see:
    *
-   * http://graphics.stanford.edu/~seander/bithacks.html#RoundUpPowerOf2
+   * https://graphics.stanford.edu/~seander/bithacks.html#RoundUpPowerOf2
    */
   --x;
   x |= x >> 1;
@@ -239,6 +239,13 @@ pocl_size_ceil2_64 (uint64_t x)
   return ++x;
 }
 
+#if defined(_WIN32) || defined(HAVE_POSIX_MEMALIGN) || defined(__ANDROID__)    \
+    || (defined(__STDC_VERSION__) && (__STDC_VERSION__ >= 201112L))
+#define HAVE_ALIGNED_ALLOC
+#else
+#error aligned malloc unavailable
+#endif
+
 static void*
 pocl_memalign_alloc(size_t align_width, size_t size)
 {
@@ -251,6 +258,10 @@ pocl_memalign_alloc(size_t align_width, size_t size)
 #elif defined(HAVE_POSIX_MEMALIGN)
   status = posix_memalign (&ptr, align_width, size);
   return ((status == 0) ? ptr : NULL);
+#elif defined(_MSC_VER)
+  return _aligned_malloc(size, align_width);
+#elif defined(__MINGW32__)
+  return __mingw_aligned_malloc(size, align_width);
 #elif (defined(__STDC_VERSION__) && (__STDC_VERSION__ >= 201112L))
   return aligned_alloc (align_width, size);
 #else
@@ -371,42 +382,110 @@ pocl_unlock_events_inorder (cl_event ev1, cl_event ev2)
     }
 }
 
-cl_int pocl_create_event (cl_event *event, cl_command_queue command_queue, 
-                          cl_command_type command_type, size_t num_buffers,
-                          const cl_mem *buffers, cl_context context)
+/* This is required because e.g. NDRange commands could have the same buffer
+ * multiple times as argument, or CopyBuffer could have src == dst buffer.
+ *
+ * If the buffer that appears multiple times in the list, is on another device,
+ * we don't want to enqueue >1 migrations for the same buffer.
+ */
+static void
+sort_and_uniq (cl_mem *objs, char *readonly_flags, size_t *num_objs)
 {
-  static unsigned int event_id_counter = 0;
+  size_t i;
+  ssize_t j;
+  size_t n = *num_objs;
+  assert (n > 1);
 
-  if (context == NULL || !(context->valid))
+  /* if the buffer is an image backed by buffer storage,
+   * replace with actual storage */
+  for (i = 0; i < n; ++i)
+    if (objs[i]->buffer)
+      objs[i] = objs[i]->buffer;
+
+  /* sort by obj id */
+  for (i = 1; i < n; ++i)
+    {
+      cl_mem buf = objs[i];
+      char c = readonly_flags[i];
+      for (j = (i - 1); ((j >= 0) && (objs[j]->id > buf->id)); --j)
+        {
+          objs[j + 1] = objs[j];
+          readonly_flags[j + 1] = readonly_flags[j];
+        }
+      objs[j + 1] = buf;
+      readonly_flags[j + 1] = c;
+    }
+
+  /* skip the first i objects which are different */
+  for (i = 1; i < n; ++i)
+    if (objs[i - 1] == objs[i])
+      break;
+
+  /* uniq */
+  size_t k = i;
+  while (i < n)
+    {
+      if (objs[k] != objs[i])
+        {
+          objs[k] = objs[i];
+          readonly_flags[k] = readonly_flags[i];
+          ++k;
+        }
+      else
+        {
+          readonly_flags[k] = readonly_flags[k] & readonly_flags[i];
+        }
+      ++i;
+    }
+
+  *num_objs = k;
+}
+
+extern unsigned long event_c;
+extern unsigned long uevent_c;
+
+cl_int
+pocl_create_event (cl_event *event, cl_command_queue command_queue,
+                   cl_command_type command_type, size_t num_buffers,
+                   const cl_mem *buffers, cl_context context)
+{
+  static uint64_t event_id_counter = 0;
+
+  if (context == NULL)
     return CL_INVALID_CONTEXT;
 
-  if (event != NULL)
+  assert (event != NULL);
+  *event = pocl_mem_manager_new_event ();
+  if (*event == NULL)
+    return CL_OUT_OF_HOST_MEMORY;
+
+  (*event)->context = context;
+  (*event)->queue = command_queue;
+
+  /* user events have a NULL command queue, don't retain it */
+  if (command_queue)
+    POname (clRetainCommandQueue) (command_queue);
+  else
+    POname (clRetainContext) (context);
+
+  (*event)->command_type = command_type;
+  (*event)->id = POCL_ATOMIC_INC (event_id_counter);
+  (*event)->num_buffers = num_buffers;
+  if (num_buffers > 0)
     {
-      *event = pocl_mem_manager_new_event ();
-      if (*event == NULL)
-        return CL_OUT_OF_HOST_MEMORY;
-
-      (*event)->context = context;
-      POname(clRetainContext) (context);
-      (*event)->queue = command_queue;
-
-      /* user events have a NULL command queue, don't retain it */
-      if (command_queue)
-        POname(clRetainCommandQueue) (command_queue);
-
-      (*event)->command_type = command_type;
-      (*event)->id = POCL_ATOMIC_INC (event_id_counter);
-      (*event)->num_buffers = num_buffers;
-
-      POCL_MSG_PRINT_EVENTS ("created event with id %d\n", (*event)->id);
-
-      if (num_buffers > 0)
-        {
-          (*event)->mem_objs = malloc (num_buffers * sizeof(cl_mem));
-          memcpy ((*event)->mem_objs, buffers, num_buffers * sizeof(cl_mem));
-        }
-      (*event)->status = CL_QUEUED;
+      (*event)->mem_objs = (cl_mem *)malloc (num_buffers * sizeof (cl_mem));
+      memcpy ((*event)->mem_objs, buffers, num_buffers * sizeof (cl_mem));
     }
+  (*event)->status = CL_QUEUED;
+
+  if (command_type == CL_COMMAND_USER)
+    POCL_ATOMIC_INC (uevent_c);
+  else
+    POCL_ATOMIC_INC (event_c);
+
+  POCL_MSG_PRINT_EVENTS ("Created event %p / ID %" PRIu64 " / Command %s\n",
+                         (*event), (*event)->id,
+                         pocl_command_to_str (command_type));
 
   return CL_SUCCESS;
 }
@@ -414,13 +493,14 @@ cl_int pocl_create_event (cl_event *event, cl_command_queue command_queue,
 static int
 pocl_create_event_sync (cl_event waiting_event, cl_event notifier_event)
 {
-  event_node * volatile notify_target = NULL;
-  event_node * volatile wait_list_item = NULL;
+  event_node *notify_target = NULL;
+  event_node *wait_list_item = NULL;
 
   if (notifier_event == NULL)
     return CL_SUCCESS;
 
-  POCL_MSG_PRINT_EVENTS ("create event sync: waiting %d, notifier %d\n",
+  POCL_MSG_PRINT_EVENTS ("create event sync: waiting %" PRIu64
+                         " , notifier %" PRIu64 "\n",
                          waiting_event->id, notifier_event->id);
 
   pocl_lock_events_inorder (waiting_event, notifier_event);
@@ -431,7 +511,10 @@ pocl_create_event_sync (cl_event waiting_event, cl_event notifier_event)
   LL_FOREACH (waiting_event->wait_list, wait_list_item)
     {
       if (wait_list_item->event == notifier_event)
-        goto FINISH;
+        {
+          POCL_MSG_PRINT_EVENTS ("Skipping event sync creation \n");
+          goto FINISH;
+        }
     }
 
   if (notifier_event->status == CL_COMPLETE)
@@ -451,39 +534,50 @@ FINISH:
   return CL_SUCCESS;
 }
 
-cl_int pocl_create_command (_cl_command_node **cmd,
+/* preallocate the buffers on destination device.
+ * if any allocation fails, we can't run this command. */
+static int
+can_run_command (cl_device_id dev, size_t num_objs, cl_mem *objs)
+{
+  size_t i;
+  int errcode;
+
+  for (i = 0; i < num_objs; ++i)
+    {
+      pocl_mem_identifier *p = &objs[i]->device_ptrs[dev->global_mem_id];
+      // skip already allocated
+      if (p->mem_ptr)
+        continue;
+
+      assert (dev->ops->alloc_mem_obj);
+      errcode = dev->ops->alloc_mem_obj (dev, objs[i], NULL);
+      if (errcode != CL_SUCCESS)
+        return CL_FALSE;
+    }
+
+  return CL_TRUE;
+}
+
+static cl_int
+pocl_create_command_struct (_cl_command_node **cmd,
                             cl_command_queue command_queue,
                             cl_command_type command_type, cl_event *event_p,
-                            cl_int num_events, const cl_event *wait_list,
+                            cl_uint num_events, const cl_event *wait_list,
                             size_t num_buffers, const cl_mem *buffers)
 {
-  int i;
+  unsigned i;
   int err;
   cl_event *event = NULL;
 
-  if ((wait_list == NULL && num_events != 0) ||
-      (wait_list != NULL && num_events == 0))
-    {
-      assert(0);
-      return CL_INVALID_EVENT_WAIT_LIST;
-    }
-
-  for (i = 0; i < num_events; ++i)
-    {
-      if (wait_list[i] == NULL)
-        return CL_INVALID_EVENT_WAIT_LIST;
-    }
-
   *cmd = pocl_mem_manager_new_command ();
   if (*cmd == NULL)
     return CL_OUT_OF_HOST_MEMORY;
 
   (*cmd)->type = command_type;
 
-  /* If user does not provide an event pointer, create an implicit one. */
   event = &((*cmd)->event);
-  err = pocl_create_event (event, command_queue, 0, num_buffers, buffers,
-                           command_queue->context);
+  err = pocl_create_event (event, command_queue, command_type, num_buffers,
+                           buffers, command_queue->context);
 
   if (err != CL_SUCCESS)
     {
@@ -514,25 +608,427 @@ cl_int pocl_create_command (_cl_command_node **cmd,
   for (i = 0; i < num_events; ++i)
     {
       cl_event wle = wait_list[i];
-      pocl_create_event_sync ((*cmd)->event, wle);
+      pocl_create_event_sync ((*event), wle);
     }
-  POCL_MSG_PRINT_EVENTS ("Created command struct (event %d, type %X)\n",
-                         (*cmd)->event->id, command_type);
+  POCL_MSG_PRINT_EVENTS (
+      "Created command struct: CMD %p (event %" PRIu64 " / %p, type: %s)\n", *cmd,
+      (*event)->id, *event, pocl_command_to_str (command_type));
   return CL_SUCCESS;
 }
 
+static int
+pocl_create_migration_commands (cl_device_id dev, cl_event final_event,
+                                cl_mem mem, pocl_mem_identifier *p,
+                                const char readonly,
+                                cl_command_type command_type,
+                                cl_mem_migration_flags mig_flags)
+{
+  int errcode = CL_SUCCESS;
+
+  cl_event ev_export = NULL, ev_import = NULL, previous_last_event = NULL,
+           last_migration_event = NULL;
+  _cl_command_node *cmd_export = NULL, *cmd_import = NULL;
+  cl_device_id ex_dev = NULL;
+  cl_command_queue ex_cq = NULL, dev_cq = NULL;
+  int can_directly_mig = 0;
+  size_t i;
+
+  /* "export" means copy buffer content from source device to mem_host_ptr;
+   *
+   * "import" means copy mem_host_ptr content to destination device,
+   * or copy directly between devices
+   *
+   * "need_hostptr" if set, increase the mem_host_ptr_refcount,
+   * to keep the mem_host_ptr backing memory around */
+  int do_import = 0, do_export = 0, do_need_hostptr = 0;
+
+  /*****************************************************************/
+
+  /* this part only:
+   *   sets up the buffer content versions according to requested migration type;
+   *   sets the buffer->last_event pointer to the final_event;
+   *   decides what needs to be actually done (import, export) but not do it;
+   *
+   * ... so that any following command sees a correct buffer state.
+   * The actual migration commands are enqueued after. */
+  POCL_LOCK_OBJ (mem);
+
+  /* Retain the buffer for the duration of the command, except Unmaps,
+   * because corresponding Maps retain twice. */
+  if (command_type != CL_COMMAND_UNMAP_MEM_OBJECT)
+    POCL_RETAIN_OBJECT_UNLOCKED (mem);
+
+  /* save buffer's current last_event as previous last_event,
+   * then set the last_event pointer to the actual command's event
+   * (final_event).
+   *
+   * We'll need the "previous" event to properly chain events, but
+   * will release it after we've enqueued the required commands. */
+  previous_last_event = mem->last_event;
+  mem->last_event = final_event;
+
+  /* find device/gmem with latest memory version and fastest migration.
+   * ex_dev = device with latest memory _other than dev_
+   * dev_cq = default command queue for destination dev */
+  int highest_d2d_mig_priority = 0;
+  for (i = 0; i < mem->context->num_devices; ++i)
+    {
+      cl_device_id d = mem->context->devices[i];
+      cl_command_queue cq = mem->context->default_queues[i];
+      if (d == dev)
+        dev_cq = cq;
+      else if (mem->device_ptrs[d->global_mem_id].version == mem->latest_version)
+        {
+          int cur_d2d_mig_priority = 0;
+          if (d->ops->can_migrate_d2d)
+            cur_d2d_mig_priority = d->ops->can_migrate_d2d (dev, d);
+
+          // if we can directly migrate, and we found a better device, use it
+          if (cur_d2d_mig_priority > highest_d2d_mig_priority)
+            {
+              ex_dev = d;
+              ex_cq = cq;
+              highest_d2d_mig_priority = cur_d2d_mig_priority;
+            }
+
+          // if we can't migrate D2D, just use plain old through-host migration
+          if (highest_d2d_mig_priority == 0)
+            {
+              ex_dev = d;
+              ex_cq = cq;
+            }
+        }
+    }
+
+  assert (dev);
+  assert (dev_cq);
+  /* ex_dev can be NULL, or non-NULL != dev */
+  assert (ex_dev != dev);
+
+  /* if mem_host_ptr_version < latest_version, one of devices must have it;
+   *
+   * could be latest_version == mem_host_ptr_version == some p->version
+   * for some p, and so i < ndev; in that case,
+   * we leave ex_dev set since D2D is preferred migration way;
+   *
+   * otherwise must be
+   * mem_host_ptr_version == latest_version & > all p->version */
+
+  if ((mem->mem_host_ptr_version < mem->latest_version) && (p->version != mem->latest_version))
+    assert ((ex_dev != NULL) && (mem->device_ptrs[ex_dev->global_mem_id].version == mem->latest_version));
+
+  /* if ex_dev is NULL, either we have the latest or it's in mem_host_ptr */
+  if (ex_dev == NULL)
+    assert ((p->version == mem->latest_version) ||
+            (mem->mem_host_ptr_version == mem->latest_version));
+
+  /*****************************************************************/
+
+  /* buffer must be already allocated on this device's globalmem */
+  assert (p->mem_ptr != NULL);
+
+  /* we're migrating to host mem only: clEnqueueMigMemObjs() with HOST flag */
+  if (mig_flags & CL_MIGRATE_MEM_OBJECT_HOST)
+    {
+      do_import = 0;
+      do_export = 0;
+      do_need_hostptr = 1;
+      if (mem->mem_host_ptr_version < mem->latest_version)
+        {
+          mem->mem_host_ptr_version = mem->latest_version;
+          /* migrate content only if needed */
+          if ((mig_flags & CL_MIGRATE_MEM_OBJECT_CONTENT_UNDEFINED) == 0)
+            {
+              /* Could be that destination dev has the latest version,
+               * we still need to migrate to host mem */
+              if (ex_dev == NULL)
+                {
+                  ex_dev = dev; ex_cq = dev_cq;
+                }
+              do_export = 1;
+              POCL_RETAIN_OBJECT_UNLOCKED (mem);
+            }
+        }
+
+      goto FINISH_VER_SETUP;
+    }
+
+  /* otherwise, we're migrating to a device memory. */
+  /* check if we can migrate to the device associated with command_queue
+   * without incurring the overhead of migrating their contents */
+  if (mig_flags & CL_MIGRATE_MEM_OBJECT_CONTENT_UNDEFINED)
+    p->version = mem->latest_version;
+
+  /* if we don't need to migrate, skip to end */
+  if (p->version >= mem->latest_version)
+    {
+      do_import = 0;
+      do_export = 0;
+      goto FINISH_VER_SETUP;
+    }
+
+  can_directly_mig = highest_d2d_mig_priority > 0;
+
+  /* if mem_host_ptr is outdated AND the devices can't migrate
+   * between each other, we need an export command */
+  if ((mem->mem_host_ptr_version != mem->latest_version)
+      && (can_directly_mig == 0))
+    {
+      /* we need two migration commands; one on the "source" device's hidden
+       * queue, and one on the destination device. */
+      do_import = 1;
+      do_export = 1;
+      do_need_hostptr = 1;
+
+      /* because the two migrate commands will clRelease the buffer */
+      POCL_RETAIN_OBJECT_UNLOCKED (mem);
+      POCL_RETAIN_OBJECT_UNLOCKED (mem);
+      mem->mem_host_ptr_version = mem->latest_version;
+      p->version = mem->latest_version;
+    }
+  /* otherwise either:
+   * 1) mem_host_ptr is latest, and we need to migrate mem-host-ptr to device, or
+   * 2) mem_host_ptr is not latest, but devices can migrate directly between each other,
+   * For both cases we only need one migration command on the destination device. */
+  else
+    {
+      do_import = 1;
+      do_export = 0;
+      do_need_hostptr = 1;
+
+      /* because the corresponding migrate command will clRelease the buffer */
+      POCL_RETAIN_OBJECT_UNLOCKED (mem);
+      p->version = mem->latest_version;
+    }
+
+FINISH_VER_SETUP:
+  /* if the command is a write-use, increase the version. */
+  if (!readonly)
+    {
+      ++p->version;
+      mem->latest_version = p->version;
+    }
+
+  if (do_need_hostptr)
+    {
+      /* increase refcount the two mig commands */
+      if (do_export)
+        ++mem->mem_host_ptr_refcount;
+      if (do_import)
+        ++mem->mem_host_ptr_refcount;
+
+      /* allocate mem_host_ptr here if needed... */
+      if (mem->mem_host_ptr == NULL)
+        {
+          size_t align = max (mem->context->min_buffer_alignment, 16);
+          mem->mem_host_ptr = pocl_aligned_malloc (align, mem->size);
+          assert ((mem->mem_host_ptr != NULL)
+                  && "Cannot allocate backing memory for mem_host_ptr!\n");
+        }
+    }
+
+  POCL_UNLOCK_OBJ (mem);
+
+  /*****************************************************************/
+
+  /* enqueue a command for export.
+   * Put the previous last event into its waitlist. */
+  if (do_export)
+    {
+      assert (ex_cq);
+      assert (ex_dev);
+      errcode = pocl_create_command_struct (
+          &cmd_export, ex_cq, CL_COMMAND_MIGRATE_MEM_OBJECTS,
+          &ev_export, // event_p
+          (previous_last_event ? 1 : 0),
+          (previous_last_event ? &previous_last_event : NULL), // waitlist
+          1, &mem                                              // buffer list
+      );
+      assert (errcode == CL_SUCCESS);
+      if (do_need_hostptr)
+        ev_export->release_mem_host_ptr_after = 1;
+
+      cmd_export->command.migrate.mem_id
+          = &mem->device_ptrs[ex_dev->global_mem_id];
+      cmd_export->command.migrate.type = ENQUEUE_MIGRATE_TYPE_D2H;
+
+      pocl_command_enqueue (ex_cq, cmd_export);
+
+      last_migration_event = ev_export;
+    }
+
+  /* enqueue a command for import.
+   * Put either the previous last event, or export ev, into its waitlist. */
+  if (do_import)
+    {
+      /* the import command must depend on (wait for) either the export
+       * command, or the buffer's previous last event. Can be NULL if there's
+       * no last event or export command */
+      cl_event import_wait_ev = (ev_export ? ev_export : previous_last_event);
+
+      errcode = pocl_create_command_struct (
+          &cmd_import, dev_cq, CL_COMMAND_MIGRATE_MEM_OBJECTS,
+          &ev_import, // event_p
+          (import_wait_ev ? 1 : 0),
+          (import_wait_ev ? &import_wait_ev : NULL), // waitlist
+          1, &mem                                    // buffer list
+      );
+      assert (errcode == CL_SUCCESS);
+      if (do_need_hostptr)
+        ev_import->release_mem_host_ptr_after = 1;
+
+      if (can_directly_mig)
+        {
+          cmd_import->command.migrate.type = ENQUEUE_MIGRATE_TYPE_D2D;
+          cmd_import->command.migrate.src_device = ex_dev;
+          cmd_import->command.migrate.src_id
+              = &mem->device_ptrs[ex_dev->global_mem_id];
+          cmd_import->command.migrate.dst_id
+              = &mem->device_ptrs[dev->global_mem_id];
+        }
+      else
+        {
+          cmd_import->command.migrate.type = ENQUEUE_MIGRATE_TYPE_H2D;
+          cmd_import->command.migrate.mem_id
+              = &mem->device_ptrs[dev->global_mem_id];
+        }
+
+      pocl_command_enqueue (dev_cq, cmd_import);
+
+      /* because explicit event */
+      if (ev_export)
+        POname (clReleaseEvent) (ev_export);
+
+      last_migration_event = ev_import;
+    }
+
+  /* we don't need it anymore. */
+  if (previous_last_event)
+    POname (clReleaseEvent (previous_last_event));
+
+  /* the final event must depend on the export/import commands */
+  if (last_migration_event)
+    {
+      pocl_create_event_sync (final_event, last_migration_event);
+      /* if the event itself only reads from the buffer,
+       * set the last buffer event to last_mig_event,
+       * instead of the actual command event;
+       * this avoids unnecessary waits e.g on kernels
+       * which only read from buffers */
+      if (readonly)
+        {
+          POCL_LOCK_OBJ (mem);
+          mem->last_event = last_migration_event;
+          POCL_UNLOCK_OBJ (mem);
+          POname (clReleaseEvent) (final_event);
+        }
+      else /* because explicit event */
+        POname (clReleaseEvent) (last_migration_event);
+    }
+
+  return CL_SUCCESS;
+}
+
+static cl_int
+pocl_create_command_full (_cl_command_node **cmd,
+                          cl_command_queue command_queue,
+                          cl_command_type command_type, cl_event *event_p,
+                          cl_uint num_events, const cl_event *wait_list,
+                          size_t num_buffers, cl_mem *buffers,
+                          char *readonly_flags,
+                          cl_mem_migration_flags mig_flags)
+{
+  cl_device_id dev = pocl_real_dev (command_queue->device);
+  int err = CL_SUCCESS;
+  size_t i;
+
+  POCL_RETURN_ERROR_ON ((dev->available == CL_FALSE), CL_INVALID_DEVICE,
+                        "device is not available\n");
+
+  if (num_buffers >= 1)
+    {
+      assert (buffers);
+      assert (readonly_flags);
+
+      if (num_buffers > 1)
+        sort_and_uniq (buffers, readonly_flags, &num_buffers);
+
+      if (can_run_command (dev, num_buffers, buffers) == CL_FALSE)
+        return CL_OUT_OF_RESOURCES;
+    }
+
+  /* waitlist here only contains the user-provided events.
+   * migration events are added to waitlist later */
+  err = pocl_create_command_struct (cmd, command_queue, command_type, event_p,
+                                    num_events, wait_list, num_buffers,
+                                    buffers);
+  if (err)
+    return err;
+  cl_event final_event = (*cmd)->event;
+
+  /* retain once for every buffer; this is because we set every buffer's
+   * "last event" to this, and then some next command enqueue
+   * (or clReleaseMemObject) will release it.
+   */
+  POCL_LOCK_OBJ (final_event);
+  final_event->pocl_refcount += num_buffers;
+  POCL_UNLOCK_OBJ (final_event);
+
+  for (i = 0; i < num_buffers; ++i)
+    {
+      pocl_create_migration_commands (
+          dev, final_event, buffers[i],
+          &buffers[i]->device_ptrs[dev->global_mem_id], readonly_flags[i],
+          command_type, mig_flags);
+    }
+
+  return err;
+}
+
+cl_int
+pocl_create_command_migrate (_cl_command_node **cmd,
+                             cl_command_queue command_queue,
+                             cl_mem_migration_flags flags, cl_event *event_p,
+                             cl_uint num_events, const cl_event *wait_list,
+                             size_t num_buffers, cl_mem *buffers,
+                             char *readonly_flags)
+{
+  return pocl_create_command_full (
+      cmd, command_queue, CL_COMMAND_MIGRATE_MEM_OBJECTS, event_p, num_events,
+      wait_list, num_buffers, buffers, readonly_flags, flags);
+}
+
+cl_int
+pocl_create_command (_cl_command_node **cmd, cl_command_queue command_queue,
+                     cl_command_type command_type, cl_event *event_p,
+                     cl_uint num_events, const cl_event *wait_list,
+                     size_t num_buffers, cl_mem *buffers, char *readonly_flags)
+{
+  return pocl_create_command_full (cmd, command_queue, command_type, event_p,
+                                   num_events, wait_list, num_buffers, buffers,
+                                   readonly_flags, 0);
+}
+
 /* call with node->event UNLOCKED */
 void pocl_command_enqueue (cl_command_queue command_queue,
                           _cl_command_node *node)
 {
   cl_event event;
 
-  POCL_LOCK_OBJ (node->event);
-  assert (node->event->status == CL_QUEUED);
-  assert (command_queue == node->event->queue);
-  POCL_UNLOCK_OBJ (node->event);
-
   POCL_LOCK_OBJ (command_queue);
+
+  /* in case of in-order queue, synchronize to previously enqueued command
+     if available */
+  if (!(command_queue->properties & CL_QUEUE_OUT_OF_ORDER_EXEC_MODE_ENABLE))
+    {
+      POCL_MSG_PRINT_EVENTS ("In-order Q; adding event syncs\n");
+      if (command_queue->last_event.event)
+        {
+          pocl_create_event_sync (node->event,
+                                  command_queue->last_event.event);
+        }
+    }
+
   ++command_queue->command_count;
   /* in case of in-order queue, synchronize to previously enqueued command
      if available */
@@ -551,11 +1047,13 @@ void pocl_command_enqueue (cl_command_queue command_queue,
             || node->type == CL_COMMAND_MARKER)
            && node->command.barrier.has_wait_list == 0)
     {
+      POCL_MSG_PRINT_EVENTS ("Barrier; adding event syncs\n");
       DL_FOREACH (command_queue->events, event)
         {
           pocl_create_event_sync (node->event, event);
         }
     }
+
   if (node->type == CL_COMMAND_BARRIER)
     command_queue->barrier = node->event;
   else
@@ -567,27 +1065,61 @@ void pocl_command_enqueue (cl_command_queue command_queue,
     }
   DL_APPEND (command_queue->events, node->event);
 
-  POCL_MSG_PRINT_EVENTS ("Last event id %d to CQ.\n", node->event->id);
+  POCL_MSG_PRINT_EVENTS ("Pushed Event %" PRIu64 " to CQ %" PRIu64 ".\n",
+                         node->event->id, command_queue->id);
   command_queue->last_event.event = node->event;
   POCL_UNLOCK_OBJ (command_queue);
 
   POCL_LOCK_OBJ (node->event);
+  assert (node->event->status == CL_QUEUED);
+  assert (command_queue == node->event->queue);
   pocl_update_event_queued (node->event);
   command_queue->device->ops->submit(node, command_queue);
   /* node->event is unlocked by device_ops->submit */
 
 }
 
+int
+pocl_alloc_or_retain_mem_host_ptr (cl_mem mem)
+{
+  if (mem->mem_host_ptr == NULL)
+    {
+      size_t align = max (mem->context->min_buffer_alignment, 16);
+      mem->mem_host_ptr = pocl_aligned_malloc (align, mem->size);
+      if (mem->mem_host_ptr == NULL)
+        return -1;
+      mem->mem_host_ptr_version = 0;
+      mem->mem_host_ptr_refcount = 0;
+    }
+  ++mem->mem_host_ptr_refcount;
+  return 0;
+}
+
+int
+pocl_release_mem_host_ptr (cl_mem mem)
+{
+  assert (mem->mem_host_ptr_refcount > 0);
+  --mem->mem_host_ptr_refcount;
+  if (mem->mem_host_ptr_refcount == 0 && mem->mem_host_ptr != NULL)
+    {
+      pocl_aligned_free (mem->mem_host_ptr);
+      mem->mem_host_ptr = NULL;
+      mem->mem_host_ptr_version = 0;
+    }
+  return 0;
+}
+
 /* call (and return) with node->event locked */
 void
-pocl_command_push (_cl_command_node *node, 
-                   _cl_command_node *volatile *ready_list, 
-                   _cl_command_node *volatile *pending_list)
+pocl_command_push (_cl_command_node *node,
+                   _cl_command_node **ready_list,
+                   _cl_command_node **pending_list)
 {
   assert (node != NULL);
 
   /* If the last command inserted is a barrier,
      command is necessary not ready */
+
   if ((*ready_list) != NULL && (*ready_list)->prev
       && (*ready_list)->prev->type == CL_COMMAND_BARRIER)
     {
@@ -605,6 +1137,30 @@ pocl_command_push (_cl_command_node *node,
     }
 }
 
+void
+pocl_unmap_command_finished (cl_device_id dev, pocl_mem_identifier *mem_id,
+                             cl_mem mem, mem_mapping_t *map)
+{
+  POCL_LOCK_OBJ (mem);
+  assert (map->unmap_requested > 0);
+  dev->ops->free_mapping_ptr (dev->data, mem_id, mem, map);
+  DL_DELETE (mem->mappings, map);
+  mem->map_count--;
+  POCL_MEM_FREE (map);
+  POCL_UNLOCK_OBJ (mem);
+}
+
+void
+pocl_unmap_command_finished2 (cl_event event, _cl_command_t *cmd)
+{
+  cl_device_id dev = event->queue->device;
+  pocl_mem_identifier *mem_id = NULL;
+  cl_mem mem = NULL;
+  mem = event->mem_objs[0];
+  mem_id = &mem->device_ptrs[dev->global_mem_id];
+  pocl_unmap_command_finished (dev, mem_id, mem, cmd->unmap.mapping);
+}
+
 void
 pocl_cl_mem_inherit_flags (cl_mem mem, cl_mem from_buffer, cl_mem_flags flags)
 {
@@ -865,7 +1421,7 @@ pocl_real_dev (const cl_device_id dev)
 cl_device_id * pocl_unique_device_list(const cl_device_id * in, cl_uint num, cl_uint *real)
 {
   cl_uint real_num = num;
-  cl_device_id * out = calloc(num, sizeof(cl_device_id));
+  cl_device_id *out = (cl_device_id *)calloc (num, sizeof (cl_device_id));
   if (!out)
     return NULL;
 
@@ -895,6 +1451,32 @@ cl_device_id * pocl_unique_device_list(const cl_device_id * in, cl_uint num, cl_
   return out;
 }
 
+int
+pocl_device_supports_builtin_kernel (cl_device_id dev, const char *kernel_name)
+{
+  if (kernel_name == NULL)
+    return 0;
+
+  if (dev->builtin_kernel_list == NULL)
+    return 0;
+
+  char *temp = strdup (dev->builtin_kernel_list);
+  char *token;
+  char *rest = temp;
+
+  while ((token = strtok_r (rest, ";", &rest)))
+    {
+      if (strcmp (token, kernel_name) == 0)
+        {
+          free (temp);
+          return 1;
+        }
+    }
+
+  free (temp);
+  return 0;
+}
+
 static void
 image_format_union (const cl_image_format *dev_formats,
                     cl_uint               num_dev_formats,
@@ -907,7 +1489,8 @@ image_format_union (const cl_image_format *dev_formats,
   if ((*num_context_formats == 0) || (*context_formats == NULL))
     {
       // alloc & copy
-      *context_formats = malloc (sizeof (cl_image_format) * num_dev_formats);
+      *context_formats = (cl_image_format *)malloc (sizeof (cl_image_format)
+                                                    * num_dev_formats);
       memcpy (*context_formats, dev_formats,
               sizeof (cl_image_format) * num_dev_formats);
       *num_context_formats = num_dev_formats;
@@ -919,7 +1502,7 @@ image_format_union (const cl_image_format *dev_formats,
       cl_uint ncf = *num_context_formats;
       size_t size = sizeof (cl_image_format) * (num_dev_formats + ncf);
       cl_image_format *ctf
-          = realloc (*context_formats, size );
+          = (cl_image_format *)realloc (*context_formats, size);
       assert (ctf);
       for (i = 0; i < num_dev_formats; ++i)
         {
@@ -948,9 +1531,11 @@ void
 pocl_setup_context (cl_context context)
 {
   unsigned i, j;
+  int err;
   size_t alignment = context->devices[0]->mem_base_addr_align;
   context->max_mem_alloc_size = 0;
   context->svm_allocdev = NULL;
+  assert (context->default_queues);
 
   memset (context->image_formats, 0, sizeof (void *) * NUM_OPENCL_IMAGE_TYPES);
   memset (context->num_image_formats, 0,
@@ -958,35 +1543,44 @@ pocl_setup_context (cl_context context)
 
   for(i=0; i<context->num_devices; i++)
     {
-      if (context->devices[i]->should_allocate_svm)
-        context->svm_allocdev = context->devices[i];
+      cl_device_id dev = context->devices[i];
+      if (dev->svm_allocation_priority > 0)
+        {
+          if (context->svm_allocdev == NULL
+              || context->svm_allocdev->svm_allocation_priority
+                     < dev->svm_allocation_priority)
+            {
+              context->svm_allocdev = dev;
+            }
+        }
 
-      if (context->devices[i]->mem_base_addr_align < alignment)
-        alignment = context->devices[i]->mem_base_addr_align;
+      if (dev->mem_base_addr_align < alignment)
+        alignment = dev->mem_base_addr_align;
 
-      if (context->devices[i]->max_mem_alloc_size
+      if (dev->max_mem_alloc_size
           > context->max_mem_alloc_size)
         context->max_mem_alloc_size =
-            context->devices[i]->max_mem_alloc_size;
+            dev->max_mem_alloc_size;
 
-      if (context->devices[i]->image_support == CL_TRUE)
+      if (dev->image_support == CL_TRUE)
         {
           for (j = 0; j < NUM_OPENCL_IMAGE_TYPES; ++j)
             image_format_union (
-                context->devices[i]->image_formats[j],
-                context->devices[i]->num_image_formats[j],
+                dev->image_formats[j],
+                dev->num_image_formats[j],
                 &context->image_formats[j], &context->num_image_formats[j]);
         }
+
+      context->default_queues[i] = POname (clCreateCommandQueue) (
+          context, dev,
+          (CL_QUEUE_OUT_OF_ORDER_EXEC_MODE_ENABLE | CL_QUEUE_HIDDEN
+           | CL_QUEUE_PROFILING_ENABLE),
+          &err);
+      assert (err == CL_SUCCESS);
+      assert (context->default_queues[i]);
     }
 
-  if (context->svm_allocdev == NULL)
-    for(i=0; i<context->num_devices; i++)
-      if (DEVICE_IS_SVM_CAPABLE(context->devices[i]))
-        {
-          context->svm_allocdev = context->devices[i];
-          break;
-        }
-
+  assert (alignment > 0);
   context->min_buffer_alignment = alignment;
 }
 
@@ -1008,7 +1602,7 @@ pocl_check_event_wait_list (cl_command_queue command_queue,
       unsigned i;
       for (i = 0; i < num_events_in_wait_list; i++)
         {
-          POCL_RETURN_ERROR_COND ((event_wait_list[i] == NULL),
+          POCL_RETURN_ERROR_COND ((!IS_CL_OBJECT_VALID (event_wait_list[i])),
                                   CL_INVALID_EVENT_WAIT_LIST);
           POCL_RETURN_ERROR_COND (
               (event_wait_list[i]->context != command_queue->context),
@@ -1134,6 +1728,25 @@ pocl_run_command (char *const *args)
   pid_t p = vfork ();
 #elif defined(HAVE_FORK)
   pid_t p = fork ();
+#elif _WIN32
+  STARTUPINFO si;
+  ZeroMemory(&si, sizeof(si));
+  si.cb = sizeof(si);
+  PROCESS_INFORMATION pi;
+  ZeroMemory(&pi, sizeof(pi));
+  DWORD dwProcessFlags = 0;
+  char * cmd = strdup(args[0]);
+  int p = CreateProcess(NULL, cmd, NULL, NULL, 1, dwProcessFlags, NULL, NULL, &si, &pi) != 0;
+  if (!p)
+    return EXIT_FAILURE;
+  DWORD waitRc = WaitForSingleObject(pi.hProcess, INFINITE);
+  if (waitRc == WAIT_FAILED)
+    return EXIT_FAILURE;
+  DWORD exit_code = 0;
+  p = GetExitCodeProcess(pi.hProcess, &exit_code) != 0;
+  if (!p)
+    return EXIT_FAILURE;
+  return exit_code;
 #else
 #error Must have fork() or vfork() system calls for HSA
 #endif
@@ -1169,6 +1782,8 @@ pocl_update_event_queued (cl_event event)
       && (cq->device->has_own_timer == 0))
     event->time_queue = pocl_gettimemono_ns ();
 
+  POCL_MSG_PRINT_EVENTS ("Event queued: %" PRIu64 "\n", event->id);
+
   if (cq->device->ops->update_event)
     cq->device->ops->update_event (cq->device, event);
   pocl_event_updated (event, CL_QUEUED);
@@ -1187,6 +1802,8 @@ pocl_update_event_submitted (cl_event event)
       && (cq->device->has_own_timer == 0))
     event->time_submit = pocl_gettimemono_ns ();
 
+  POCL_MSG_PRINT_EVENTS ("Event submitted: %" PRIu64 "\n", event->id);
+
   if (cq->device->ops->update_event)
     cq->device->ops->update_event (cq->device, event);
   pocl_event_updated (event, CL_SUBMITTED);
@@ -1204,6 +1821,8 @@ pocl_update_event_running_unlocked (cl_event event)
       && (cq->device->has_own_timer == 0))
     event->time_start = pocl_gettimemono_ns ();
 
+  POCL_MSG_PRINT_EVENTS ("Event running: %" PRIu64 "\n", event->id);
+
   if (cq->device->ops->update_event)
     cq->device->ops->update_event (cq->device, event);
   pocl_event_updated (event, CL_RUNNING);
@@ -1239,16 +1858,14 @@ pocl_update_event_finished_msg (cl_int status, const char *func, unsigned line,
     ops->update_event (cq->device, event);
 
   if (status == CL_COMPLETE)
-    POCL_MSG_PRINT_EVENTS ("%s: Command complete, event %d\n",
+    POCL_MSG_PRINT_EVENTS ("%s: Command complete, event %" PRIu64 "\n",
                            cq->device->short_name, event->id);
   else
-    POCL_MSG_PRINT_EVENTS ("%s: Command FAILED, event %d\n",
+    POCL_MSG_PRINT_EVENTS ("%s: Command FAILED, event %" PRIu64 "\n",
                            cq->device->short_name, event->id);
 
-  pocl_mem_objs_cleanup (event);
-
+  assert (cq->command_count > 0);
   --cq->command_count;
-  assert (cq->command_count >= 0);
   if (cq->barrier == event)
     cq->barrier = NULL;
   if (cq->last_event.event == event)
@@ -1277,6 +1894,20 @@ pocl_update_event_finished_msg (cl_int status, const char *func, unsigned line,
     }
 #endif
 
+  size_t i;
+  for (i = 0; i < event->num_buffers; ++i)
+    {
+      cl_mem mem = event->mem_objs[i];
+      if (event->release_mem_host_ptr_after)
+        {
+          POCL_LOCK_OBJ (mem);
+          pocl_release_mem_host_ptr (mem);
+          POCL_UNLOCK_OBJ (mem);
+        }
+      POname (clReleaseMemObject) (mem);
+    }
+  POCL_MEM_FREE (event->mem_objs);
+
   POname (clReleaseEvent) (event);
 }
 
@@ -1376,3 +2007,43 @@ float_to_half (float value)
   v.si ^= ((v.si - minD) ^ v.si) & -(v.si > subC);
   return v.ui | sign;
 }
+
+/* SPIR-V magic header */
+#define SPIRV_MAGIC 0x07230203U
+/* Opcode for capability used by module */
+#define OpCapab 0x00020011
+/* execution model = Kernel is used by OpenCL SPIR-V modules */
+#define KernelExecModel 0x6
+
+int
+bitcode_is_spirv_kernel (const char *bitcode, size_t size)
+{
+  const uint32_t *bc32 = (const uint32_t *)bitcode;
+  unsigned location = 0;
+  uint32_t header_magic = htole32 (bc32[location++]);
+
+  if ((size < 20) || (header_magic != SPIRV_MAGIC))
+    return 0;
+
+  // skip version, generator, bound, schema
+  location += 4;
+  int is_opencl = 0;
+  uint32_t instruction, value;
+  do
+    {
+      instruction = htole32 (bc32[location++]);
+      value = htole32 (bc32[location++]);
+      if (value == KernelExecModel)
+        is_opencl = 1;
+    }
+  while (instruction == OpCapab);
+
+  /* SPIR-V but not OpenCL-type. */
+  if (!is_opencl)
+    {
+      POCL_MSG_ERR ("SPIR-V binary provided, but is not using Kernel mode."
+                    "Pocl can't process this binary.\n");
+    }
+
+  return is_opencl;
+}
