@@ -149,18 +149,24 @@ WorkItemAliasAnalysis::getAnalysisUsage(AnalysisUsage &AU) const {
  * Then they can not alias.
  */
 
+#ifdef LLVM_OLDER_THAN_13_0
+#define NO_ALIAS NoAlias
+#else
+#define NO_ALIAS AliasResult::Kind::NoAlias
+#endif
+
 AliasResult
 WorkItemAAResult::alias(const Location &LocA, const Location &LocB) {
     // If either of the memory references is empty, it doesn't matter what the
     // pointer values are. This allows the code below to ignore this special
     // case.
     if (LocA.Size == 0 || LocB.Size == 0)
-        return NoAlias;
+      return NO_ALIAS;
 
     // Pointers from different address spaces do not alias
     if (cast<PointerType>(LocA.Ptr->getType())->getAddressSpace() != 
         cast<PointerType>(LocB.Ptr->getType())->getAddressSpace()) {
-        return NoAlias;
+      return NO_ALIAS;
     }
     // In case code is created by pocl, we can also use metadata.
     if (isa<Instruction>(LocA.Ptr) && isa<Instruction>(LocB.Ptr)) {
@@ -217,7 +223,7 @@ WorkItemAAResult::alias(const Location &LocA, const Location &LocB) {
                 if ( !(CIX->getValue() == CJX->getValue()
                     && CIY->getValue() == CJY->getValue()
                     && CIZ->getValue() == CJZ->getValue())) {
-                    return NoAlias;
+                  return NO_ALIAS;
                 }                
             }        
         }
