@@ -31,7 +31,7 @@ IGNORE_COMPILER_WARNING("-Wunused-parameter")
 
 #include "config.h"
 #include "pocl.h"
-#include "pocl_cl.h"
+#include "pocl_llvm_api.h"
 
 #include <llvm/Analysis/ConstantFolding.h>
 #include <llvm/IR/BasicBlock.h>
@@ -59,11 +59,12 @@ IGNORE_COMPILER_WARNING("-Wunused-parameter")
 #include <llvm-c/Core.h>
 #include <llvm-c/Target.h>
 
-#include "CanonicalizeBarriers.h"
-#include "BarrierTailReplication.h"
-#include "WorkitemReplication.h"
 #include "Barrier.h"
+#include "BarrierTailReplication.h"
+#include "CanonicalizeBarriers.h"
+#include "LLVMUtils.h"
 #include "Workgroup.h"
+#include "WorkitemReplication.h"
 
 #include <cstdio>
 #include <map>
@@ -78,15 +79,11 @@ IGNORE_COMPILER_WARNING("-Wunused-parameter")
 
 POP_COMPILER_DIAGS
 
-extern cl_device_id currentPoclDevice;
-
 using namespace std;
 using namespace llvm;
 using namespace pocl;
 
-/* The kernel to process in this kernel compiler launch. */
-cl::opt<string> KernelName("kernel", cl::desc("Kernel function name"),
-                           cl::value_desc("kernel"), cl::init(""));
+
 
 #ifdef LLVM_OLDER_THAN_7_0
 namespace llvm {
@@ -170,8 +167,36 @@ Workgroup::runOnModule(Module &M) {
   this->M = &M;
   this->C = &M.getContext();
 
+  getModuleIntMetadata(M, "device_address_bits", address_bits);
+  getModuleBoolMetadata(M, "device_arg_buffer_launcher",
+                        DeviceUsingArgBufferLauncher);
+  getModuleBoolMetadata(M, "device_is_spmd", DeviceIsSPMD);
+
+  getModuleStringMetadata(M, "KernelName", KernelName);
+  getModuleIntMetadata(M, "WGMaxGridDimWidth", WGMaxGridDimWidth);
+  getModuleIntMetadata(M, "WGLocalSizeX", WGLocalSizeX);
+  getModuleIntMetadata(M, "WGLocalSizeY", WGLocalSizeY);
+  getModuleIntMetadata(M, "WGLocalSizeZ", WGLocalSizeZ);
+  getModuleBoolMetadata(M, "WGDynamicLocalSize", WGDynamicLocalSize);
+  getModuleBoolMetadata(M, "WGAssumeZeroGlobalOffset",
+                        WGAssumeZeroGlobalOffset);
+
+  getModuleIntMetadata(M, "device_global_as_id", DeviceGlobalASid);
+  getModuleIntMetadata(M, "device_local_as_id", DeviceLocalASid);
+  getModuleIntMetadata(M, "device_constant_as_id", DeviceConstantASid);
+  getModuleIntMetadata(M, "device_args_as_id", DeviceArgsASid);
+  getModuleIntMetadata(M, "device_context_as_id", DeviceContextASid);
+
+  getModuleBoolMetadata(M, "device_side_printf", DeviceSidePrintf);
+  getModuleBoolMetadata(M, "device_alloca_locals", DeviceAllocaLocals);
+
+  getModuleIntMetadata(M, "device_max_witem_dim", DeviceMaxWItemDim);
+  getModuleIntMetadata(M, "device_max_witem_sizes_0", DeviceMaxWItemSizes[0]);
+  getModuleIntMetadata(M, "device_max_witem_sizes_1", DeviceMaxWItemSizes[1]);
+  getModuleIntMetadata(M, "device_max_witem_sizes_2", DeviceMaxWItemSizes[2]);
+
   HiddenArgs = 0;
-  SizeTWidth = currentPoclDevice->address_bits;
+  SizeTWidth = address_bits;
   SizeT = IntegerType::get(*C, SizeTWidth);
 
 #ifdef LLVM_OLDER_THAN_7_0
@@ -203,9 +228,8 @@ Workgroup::runOnModule(Module &M) {
   LauncherFuncT = FunctionType::get(
       Type::getVoidTy(*C),
       {PointerType::get(PointerType::get(Type::getInt8Ty(*C), 0),
-                        currentPoclDevice->args_as_id),
-       PointerType::get(PoclContextT, currentPoclDevice->context_as_id), SizeT,
-       SizeT, SizeT},
+                        DeviceArgsASid),
+       PointerType::get(PoclContextT, DeviceContextASid), SizeT, SizeT, SizeT},
       false);
 #endif
 
@@ -237,14 +261,14 @@ Workgroup::runOnModule(Module &M) {
 
     privatizeContext(L);
 
-    if (currentPoclDevice->arg_buffer_launcher) {
+    if (DeviceUsingArgBufferLauncher) {
       Function *WGLauncher =
         createArgBufferWorkgroupLauncher(L, OrigKernel.getName().str());
       L->addFnAttr(Attribute::NoInline);
       L->removeFnAttr(Attribute::AlwaysInline);
       WGLauncher->addFnAttr(Attribute::AlwaysInline);
       createGridLauncher(L, WGLauncher, OrigKernel.getName().str());
-    } else if (currentPoclDevice->spmd) {
+    } else if (DeviceIsSPMD) {
       // For SPMD machines there is no need for a WG launcher, the device will
       // call/handle the single-WI kernel function directly.
       kernels[&OrigKernel] = L;
@@ -256,7 +280,7 @@ Workgroup::runOnModule(Module &M) {
     }
   }
 
-  if (!currentPoclDevice->arg_buffer_launcher && currentPoclDevice->spmd) {
+  if (!DeviceUsingArgBufferLauncher && DeviceIsSPMD) {
     regenerate_kernel_metadata(M, kernels);
 
     // Delete the old kernels.
@@ -297,16 +321,16 @@ static void addRangeMetadata(llvm::Instruction *Instr, size_t Min, size_t Max) {
   Instr->setMetadata(LLVMContext::MD_range, Range);
 }
 
-static void addRangeMetadataForPCField(llvm::Instruction *Instr,
-                                       int StructFieldIndex,
-                                       int FieldIndex = -1) {
+void Workgroup::addRangeMetadataForPCField(llvm::Instruction *Instr,
+                                           int StructFieldIndex,
+                                           int FieldIndex) {
   uint64_t Min = 0;
   uint64_t Max = 0;
   uint64_t LocalSizes[] = {WGLocalSizeX, WGLocalSizeY, WGLocalSizeZ};
   switch (StructFieldIndex) {
   case PC_WORK_DIM:
     Min = 1;
-    Max = currentPoclDevice->max_work_item_dimensions;
+    Max = DeviceMaxWItemDim;
     break;
   case PC_NUM_GROUPS:
     Min = 1;
@@ -348,8 +372,7 @@ static void addRangeMetadataForPCField(llvm::Instruction *Instr,
       if (WGDynamicLocalSize) {
         Max = (WGMaxGridDimWidth > 0
                    ? WGMaxGridDimWidth
-                   : min(currentPoclDevice->max_work_item_sizes[FieldIndex],
-                         WGMaxGridDimWidth));
+                   : min(DeviceMaxWItemSizes[FieldIndex], WGMaxGridDimWidth));
       } else {
         // The local size is converted to constant with static WGs, so this is
         // actually useless.
@@ -380,21 +403,35 @@ Workgroup::createLoadFromContext(
   IRBuilder<> &Builder, int StructFieldIndex, int FieldIndex=-1) {
 
   Value *GEP;
-  GEP = Builder.CreateStructGEP(ContextArg->getType()->getPointerElementType(),
-                                ContextArg, StructFieldIndex);
+  Type *ContextType = ContextArg->getType()->getPointerElementType();
+  GEP = Builder.CreateStructGEP(ContextType, ContextArg, StructFieldIndex);
+  Type *GEPType = GEP->getType()->getPointerElementType();
 
   llvm::LoadInst *Load = nullptr;
   if (SizeTWidth == 64) {
     if (FieldIndex == -1)
-      Load = Builder.CreateLoad(Builder.CreateConstGEP1_64(GEP, 0));
+      Load = Builder.CreateLoad(Builder.CreateConstGEP1_64(
+#ifndef LLVM_OLDER_THAN_13_0
+          GEPType,
+#endif
+          GEP, 0));
     else
-      Load = Builder.CreateLoad(Builder.CreateConstGEP2_64(GEP, 0, FieldIndex));
+      Load = Builder.CreateLoad(Builder.CreateConstGEP2_64(
+#ifndef LLVM_OLDER_THAN_13_0
+          GEPType,
+#endif
+          GEP, 0, FieldIndex));
   } else {
     if (FieldIndex == -1)
-      Load = Builder.CreateLoad(Builder.CreateConstGEP1_32(GEP, 0));
+      Load = Builder.CreateLoad(Builder.CreateConstGEP1_32(
+#ifndef LLVM_OLDER_THAN_13_0
+          GEPType,
+#endif
+          GEP, 0));
     else
       Load = Builder.CreateLoad(Builder.CreateConstGEP2_32(
-          GEP->getType()->getPointerElementType(), GEP, 0, FieldIndex));
+          GEPType,
+          GEP, 0, FieldIndex));
   }
   addRangeMetadataForPCField(Load, StructFieldIndex, FieldIndex);
   return Load;
@@ -466,7 +503,7 @@ static Function *cloneFunctionWithPrintfArgs(Value *pb, Value *pbp, Value *pbc,
   // otherwise there will be an assertion. The changes are likely
   // additional debug info nodes added when cloning the function into
   // the other.  For some reason it doesn't want to reuse the old ones.
-  CloneFunctionInto(NewF, F, VV, true, RI);
+  CloneFunctionIntoAbs(NewF, F, VV, RI);
 
   return NewF;
 }
@@ -619,14 +656,12 @@ Workgroup::createWrapper(Function *F, FunctionMapping &printfCache) {
        i != e; ++i)
     sv.push_back(i->getType());
 
-  if (!currentPoclDevice->arg_buffer_launcher && currentPoclDevice->spmd) {
-    sv.push_back(
-      PointerType::get(PoclContextT, currentPoclDevice->context_as_id));
+  if (!DeviceUsingArgBufferLauncher && DeviceIsSPMD) {
+    sv.push_back(PointerType::get(PoclContextT, DeviceContextASid));
     HiddenArgs = 1;
   } else {
     // pocl_context
-    sv.push_back(
-      PointerType::get(PoclContextT, currentPoclDevice->context_as_id));
+    sv.push_back(PointerType::get(PoclContextT, DeviceContextASid));
     // group_x
     sv.push_back(SizeT);
     // group_y
@@ -646,7 +681,7 @@ Workgroup::createWrapper(Function *F, FunctionMapping &printfCache) {
   std::string funcName = "";
   funcName = F->getName().str();
   Function *L = NULL;
-  if (!currentPoclDevice->arg_buffer_launcher && currentPoclDevice->spmd) {
+  if (!DeviceUsingArgBufferLauncher && DeviceIsSPMD) {
     Function *F = M->getFunction(funcName);
     F->setName(funcName + "_original");
     L = Function::Create(ft, Function::ExternalLinkage, funcName, M);
@@ -683,7 +718,7 @@ Workgroup::createWrapper(Function *F, FunctionMapping &printfCache) {
   IRBuilder<> Builder(BasicBlock::Create(C, "", L));
 
   Value *pb, *pbp, *pbc;
-  if (currentPoclDevice->device_side_printf) {
+  if (DeviceSidePrintf) {
     pb = createLoadFromContext(Builder, PC_PRINTF_BUFFER);
     pbp = createLoadFromContext(Builder, PC_PRINTF_BUFFER_POSITION);
     pbc = createLoadFromContext(Builder, PC_PRINTF_BUFFER_CAPACITY);
@@ -737,7 +772,7 @@ Workgroup::createWrapper(Function *F, FunctionMapping &printfCache) {
   InlineFunction(*c, IFI);
 #endif
 
-  if (currentPoclDevice->device_side_printf) {
+  if (DeviceSidePrintf) {
     Function *poclPrintf = M->getFunction("__pocl_printf");
     replacePrintfCalls(pb, pbp, pbc, true, poclPrintf, *M, L, printfCache);
   }
@@ -748,7 +783,7 @@ Workgroup::createWrapper(Function *F, FunctionMapping &printfCache) {
   // kernels that should be executed in SPMD fashion. For MIMD/CPU,
   // we want to use the default calling convention for the work group
   // function.
-  if (currentPoclDevice->spmd)
+  if (DeviceIsSPMD)
     L->setCallingConv(F->getCallingConv());
 
   return L;
@@ -934,7 +969,7 @@ Workgroup::privatizeContext(Function *F)
       Builder, {"_num_groups_x", "_num_groups_y", "_num_groups_z"},
       PC_NUM_GROUPS));
 
-  if (currentPoclDevice->device_side_printf) {
+  if (DeviceSidePrintf) {
     // Privatize _printf_buffer
     privatizeGlobals(
       F, Builder, {"_printf_buffer"}, {
@@ -988,14 +1023,14 @@ Workgroup::createDefaultWorkgroupLauncher(llvm::Function *F) {
       break;
 
     Type *ArgType = ii->getType();
-
-    Value *GEP = Builder.CreateGEP(
-        &*ai, ConstantInt::get(IntegerType::get(M->getContext(), 32), i));
-    Value *Pointer = Builder.CreateLoad(GEP);
+    Type* I32Ty = Type::getInt32Ty(M->getContext());
+    Value *AI = &*ai;
+    Value *GEP = Builder.CreateGEP(AI->getType()->getPointerElementType(),
+        AI, ConstantInt::get(I32Ty, i));
+    Value *Pointer = Builder.CreateLoad(GEP->getType()->getPointerElementType(), GEP);
 
     Value *Arg;
-    if (currentPoclDevice->device_alloca_locals &&
-        isLocalMemFunctionArg(F, i)) {
+    if (DeviceAllocaLocals && isLocalMemFunctionArg(F, i)) {
       // Generate allocas for the local buffer arguments.
       PointerType *ParamType = dyn_cast<PointerType>(ArgType);
       Type *ArgElementType = ParamType->getElementType();
@@ -1052,7 +1087,7 @@ Workgroup::createDefaultWorkgroupLauncher(llvm::Function *F) {
         Arg = Builder.CreatePointerCast(Pointer, ArgType);
       } else {
         Arg = Builder.CreatePointerCast(Pointer, ArgType->getPointerTo());
-        Arg = Builder.CreateLoad(Arg);
+        Arg = Builder.CreateLoad(ArgType, Arg);
       }
     }
 
@@ -1100,6 +1135,22 @@ static size_t getArgumentSize(llvm::Argument &Arg) {
   return DL.getTypeStoreSize(TypeInBuf);
 }
 
+static uint64_t pocl_size_ceil2_64(uint64_t x) {
+  /* Rounds up to the next highest power of two without branching and
+   * is as fast as a BSR instruction on x86, see:
+   *
+   * http://graphics.stanford.edu/~seander/bithacks.html#RoundUpPowerOf2
+   */
+  --x;
+  x |= x >> 1;
+  x |= x >> 2;
+  x |= x >> 4;
+  x |= x >> 8;
+  x |= x >> 16;
+  x |= x >> 32;
+  return ++x;
+}
+
 static void computeArgBufferOffsets(LLVMValueRef F,
                                     uint64_t *ArgBufferOffsets) {
 
@@ -1112,19 +1163,75 @@ static void computeArgBufferOffsets(LLVMValueRef F,
     // TODO: This is a target specific type? We would like to get the
     // natural size or the "packed size" instead...
     uint64_t ByteSize = getArgumentSize(cast<Argument>(*unwrap(Param)));
-    uint64_t Alignment = ByteSize;
+    uint64_t Alignment = pocl_size_ceil2_64(ByteSize);
 
     assert(ByteSize > 0 && "Arg type size is zero?");
     Offset = align64(Offset, Alignment);
+
     ArgBufferOffsets[i] = Offset;
     Offset += ByteSize;
   }
 }
 
-static LLVMValueRef
-createArgBufferLoad(LLVMBuilderRef Builder, LLVMValueRef ArgBufferPtr,
-                    uint64_t *ArgBufferOffsets, LLVMValueRef F,
-                    unsigned ParamIndex) {
+LLVMValueRef
+Workgroup::createAllocaMemcpyForStruct(LLVMModuleRef M, LLVMBuilderRef Builder,
+                                       llvm::Argument &Arg,
+                                       LLVMValueRef ArgByteOffset) {
+
+  LLVMContextRef LLVMContext = LLVMGetModuleContext(M);
+  LLVMValueRef MemCpy1 = LLVMGetNamedFunction(M, "_pocl_memcpy_1");
+  LLVMValueRef MemCpy4 = LLVMGetNamedFunction(M, "_pocl_memcpy_4");
+  LLVMTypeRef Int8Type = LLVMInt8TypeInContext(LLVMContext);
+  LLVMTypeRef Int32Type = LLVMInt32TypeInContext(LLVMContext);
+
+  llvm::Type *TypeInArg = Arg.getType()->getPointerElementType();
+  const DataLayout &DL = Arg.getParent()->getParent()->getDataLayout();
+  unsigned alignment = DL.getABITypeAlignment(TypeInArg);
+  uint64_t StoreSize = DL.getTypeStoreSize(TypeInArg);
+  LLVMValueRef Size =
+      LLVMConstInt(LLVMInt32TypeInContext(LLVMContext), StoreSize, 0);
+
+  LLVMValueRef LocalArgAlloca =
+      LLVMBuildAlloca(Builder, wrap(TypeInArg), "struct_arg");
+
+  if ((alignment % 4 == 0) && (StoreSize % 4 == 0)) {
+    LLVMTypeRef i32PtrAS0 = LLVMPointerType(Int32Type, 0);
+    LLVMTypeRef i32PtrAS1 = LLVMPointerType(Int32Type, DeviceArgsASid);
+    LLVMValueRef CARG0 =
+        LLVMBuildPointerCast(Builder, LocalArgAlloca, i32PtrAS0, "cargDst");
+    LLVMValueRef CARG1 =
+        LLVMBuildPointerCast(Builder, ArgByteOffset, i32PtrAS1, "cargSrc");
+
+    LLVMValueRef args[3];
+    args[0] = CARG0;
+    args[1] = CARG1;
+    args[2] = Size;
+
+    LLVMValueRef call4 = LLVMBuildCall(Builder, MemCpy4, args, 3, "");
+  } else {
+    LLVMTypeRef i8PtrAS0 = LLVMPointerType(Int8Type, 0);
+    LLVMTypeRef i8PtrAS1 = LLVMPointerType(Int8Type, DeviceArgsASid);
+    LLVMValueRef CARG0 =
+        LLVMBuildPointerCast(Builder, LocalArgAlloca, i8PtrAS0, "cargDst");
+    LLVMValueRef CARG1 =
+        LLVMBuildPointerCast(Builder, ArgByteOffset, i8PtrAS1, "cargSrc");
+
+    LLVMValueRef args[3];
+    args[0] = CARG0;
+    args[1] = CARG1;
+    args[2] = Size;
+
+    LLVMValueRef call1 = LLVMBuildCall(Builder, MemCpy1, args, 3, "");
+  }
+
+  return LocalArgAlloca;
+}
+
+LLVMValueRef Workgroup::createArgBufferLoad(LLVMBuilderRef Builder,
+                                            LLVMValueRef ArgBufferPtr,
+                                            uint64_t *ArgBufferOffsets,
+                                            LLVMValueRef F,
+                                            unsigned ParamIndex) {
 
   LLVMValueRef Param = LLVMGetParam(F, ParamIndex);
   LLVMTypeRef ParamType = LLVMTypeOf(Param);
@@ -1138,14 +1245,26 @@ createArgBufferLoad(LLVMBuilderRef Builder, LLVMValueRef ArgBufferPtr,
   LLVMValueRef ArgByteOffset =
       LLVMBuildGEP(Builder, ArgBufferPtr, &Offs, 1, "arg_byte_offset");
 
-  if (isByValPtrArgument(cast<Argument>(*unwrap(Param)))) {
-    // In case of byval arguments (private structs), the struct
-    // is in the arg buffer directly. Just refer to its address.
-    return LLVMBuildPointerCast(Builder, ArgByteOffset, ParamType,
-                                "inval_arg_ptr");
+  llvm::Argument &Arg = cast<Argument>(*unwrap(Param));
+
+  // byval arguments (private structs), passed via pointer
+  if (isByValPtrArgument(Arg)) {
+
+    // the kernel AS for private structs is always zero (private).
+    // if the arg address space is also zero, nothing to do here just cast...
+    if (DeviceArgsASid == 0)
+      return LLVMBuildPointerCast(Builder, ArgByteOffset, ParamType,
+                                  "inval_arg_ptr");
+
+    // ... otherwise the arg AS is different, and we need an alloca+memcpy.
+    else
+      return createAllocaMemcpyForStruct(M, Builder, Arg, ArgByteOffset);
+
+    // not by-val argument
   } else {
     LLVMValueRef ArgOffsetBitcast = LLVMBuildPointerCast(
-        Builder, ArgByteOffset, LLVMPointerType(ParamType, 0), "arg_ptr");
+        Builder, ArgByteOffset, LLVMPointerType(ParamType, DeviceArgsASid),
+        "arg_ptr");
     return LLVMBuildLoad(Builder, ArgOffsetBitcast, "");
   }
 }
@@ -1176,11 +1295,9 @@ Workgroup::createArgBufferWorkgroupLauncher(Function *Func,
   LLVMTypeRef Int32Type = LLVMInt32TypeInContext(LLVMContext);
   LLVMTypeRef Int64Type = LLVMInt64TypeInContext(LLVMContext);
 
-  LLVMTypeRef ArgsPtrType =
-    LLVMPointerType(Int8Type, currentPoclDevice->args_as_id);
+  LLVMTypeRef ArgsPtrType = LLVMPointerType(Int8Type, DeviceArgsASid);
 
-  LLVMTypeRef CtxPtrType =
-    LLVMPointerType(Int8Type, currentPoclDevice->context_as_id);
+  LLVMTypeRef CtxPtrType = LLVMPointerType(Int8Type, DeviceContextASid);
 
   std::ostringstream StrStr;
   StrStr << KernName;
@@ -1215,10 +1332,8 @@ Workgroup::createArgBufferWorkgroupLauncher(Function *Func,
   LLVMValueRef Args[ArgCount];
   LLVMValueRef ArgBuffer = LLVMGetParam(WrapperKernel, 0);
   size_t i = 0;
-  for (; i < ArgCount - HiddenArgs + 1; ++i) {
-
-    if (currentPoclDevice->device_alloca_locals &&
-        isLocalMemFunctionArg(Func, i)) {
+  for (; i < ArgCount - HiddenArgs; ++i) {
+    if (DeviceAllocaLocals && isLocalMemFunctionArg(Func, i)) {
 
       // Generate allocas for the local buffer arguments.
 
@@ -1295,7 +1410,14 @@ Workgroup::createArgBufferWorkgroupLauncher(Function *Func,
     }
   }
 
-  size_t Arg = 2;
+  size_t Arg = 1;
+  // Pass the context object
+  LLVMValueRef CtxParam = LLVMGetParam(WrapperKernel, Arg++);
+  LLVMTypeRef CtxT = wrap(PoclContextT);
+  LLVMTypeRef CtxPtrTypeActual = LLVMPointerType(CtxT, DeviceContextASid);
+  LLVMValueRef CastContext =
+      LLVMBuildPointerCast(Builder, CtxParam, CtxPtrTypeActual, "ctx_ptr");
+  Args[i++] = CastContext;
   // Pass the group ids.
   Args[i++] = LLVMGetParam(WrapperKernel, Arg++);
   Args[i++] = LLVMGetParam(WrapperKernel, Arg++);
@@ -1303,13 +1425,13 @@ Workgroup::createArgBufferWorkgroupLauncher(Function *Func,
 
   assert (i == ArgCount);
 
-  // Pass the context object.
   LLVMValueRef Call = LLVMBuildCall(Builder, F, Args, ArgCount, "");
   LLVMBuildRetVoid(Builder);
 
   llvm::CallInst *CallI = llvm::dyn_cast<llvm::CallInst>(llvm::unwrap(Call));
   CallI->setCallingConv(Func->getCallingConv());
 
+  LLVMDisposeBuilder(Builder);
   return llvm::dyn_cast<llvm::Function>(llvm::unwrap(WrapperKernel));
 }
 
@@ -1332,8 +1454,7 @@ Workgroup::createGridLauncher(Function *KernFunc, Function *WGFunc,
 
   LLVMTypeRef Int8Type = LLVMInt8TypeInContext(LLVMContext);
   LLVMTypeRef Int8PtrType = LLVMPointerType(Int8Type, 0);
-  LLVMTypeRef ArgsPtrType =
-      LLVMPointerType(Int8Type, currentPoclDevice->args_as_id);
+  LLVMTypeRef ArgsPtrType = LLVMPointerType(Int8Type, DeviceArgsASid);
 
   std::ostringstream StrStr("phsa_kernel.", std::ios::ate);
   StrStr << KernName;
@@ -1445,38 +1566,42 @@ Workgroup::createFastWorkgroupLauncher(llvm::Function *F) {
     if (i == F->arg_size() - 4)
       break;
 
-    Type *t = ii->getType();
-    Value *gep = builder.CreateGEP(&*ai,
-            ConstantInt::get(IntegerType::get(M->getContext(), 32), i));
-    Value *pointer = builder.CreateLoad(gep);
+    Type *T = ii->getType();
+    Type* I32Ty = Type::getInt32Ty(M->getContext());
+    Value *AI = &*ai;
+    Value *GEP = builder.CreateGEP(AI->getType()->getPointerElementType(),
+        AI, ConstantInt::get(I32Ty, i));
+    Value *Pointer = builder.CreateLoad(GEP->getType()->getPointerElementType(), GEP);
+    Value *V;
 
-    if (t->isPointerTy()) {
+    if (T->isPointerTy()) {
       if (!ii->hasByValAttr()) {
         // Assume the pointer is directly in the arg array.
-        arguments.push_back(builder.CreatePointerCast(pointer, t));
+        V = builder.CreatePointerCast(Pointer, T);
+        arguments.push_back(V);
         continue;
+      } else {
+        // It's a pass by value pointer argument, use the underlying
+        // element type in subsequent load.
+        T = T->getPointerElementType();
       }
-
-      // It's a pass by value pointer argument, use the underlying
-      // element type in subsequent load.
-      t = t->getPointerElementType();
     }
 
     // If it's a pass by value pointer argument, we just pass the pointer
     // as is to the function, no need to load from it first.
-    Value *value;
 
-    if (!ii->hasByValAttr() || ((PointerType*)t)->getAddressSpace() == 1)
-      value = builder.CreatePointerCast
-        (pointer, t->getPointerTo(currentPoclDevice->global_as_id));
-    else
-      value = builder.CreatePointerCast(pointer, t->getPointerTo());
-
-    if (!ii->hasByValAttr()) {
-      value = builder.CreateLoad(value);
+    if (ii->hasByValAttr() && (((PointerType *)T)->getAddressSpace() != DeviceGlobalASid)) {
+      V = builder.CreatePointerCast(Pointer, T->getPointerTo());
+    } else {
+      V =
+          builder.CreatePointerCast(Pointer, T->getPointerTo(DeviceGlobalASid));
     }
 
-    arguments.push_back(value);
+    if (!ii->hasByValAttr()) {
+      V = builder.CreateLoad(T, V);
+    }
+
+    arguments.push_back(V);
   }
 
   ++ai;
@@ -1504,9 +1629,13 @@ bool Workgroup::isKernelToProcess(const Function &F) {
 
   NamedMDNode *kernels = m->getNamedMetadata("opencl.kernels");
   if (kernels == NULL) {
+
+    std::string KernelName;
+    getModuleStringMetadata(*m, "KernelName", KernelName);
+
     if (KernelName == "")
       return true;
-    if (F.getName() == KernelName)
+    if (F.getName().str() == KernelName)
       return true;
 
     return false;
