@@ -147,16 +147,15 @@ main(void)
             cBuffer,
             (void *) output);
 
-        // There is no need to perform a finish on the final unmap
-        // or release any objects as this all happens implicitly with
-        // the C++ Wrapper API.
+        queue.finish();
+        platformList[0].unloadCompiler();
 
         if (ok)
           return EXIT_SUCCESS;
         else
           return EXIT_FAILURE;
     }
-    catch (cl::Error err) {
+    catch (cl::Error &err) {
          std::cerr
              << "ERROR: "
              << err.what()
