@@ -32,10 +32,11 @@ int main(void)
 
 	result[rvs]=0;	// spec doesn't say it is null-terminated.
         const char *expected = "OpenCL " POCL_CL_VERSION
-                               " pocl " PACKAGE_VERSION;
+                               " pocl " POCL_VERSION_FULL;
         if (strncmp (result, expected, strlen (expected)) != 0)
           {
             printf ("Error: platform is: %s\n", result);
+            printf ("Should be: %s\n", expected);
             return 2;
           }
 
